#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   k8s_config.py
@Author  :   Wei Loon Cheng, Kai Xuan Lee
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Custom runner config functions.
"""

import os
import uuid
from pathlib import Path

from datature.nexus.cli.runner import messages as MESSAGES
from datature.nexus.cli.runner.common import install_pip_package
from datature.nexus.cli.runner.config import RunnerConfig
from datature.nexus.cli.runner.consts import (
    RUNNER_DEPLOYMENT_CONFIG_NAME,
    RUNNER_DEPLOYMENT_NAME,
)
from datature.nexus.error import Error

try:
    from kubernetes import client as k8s_client
except ImportError as exc:
    if not install_pip_package("kubernetes"):
        raise Error(
            MESSAGES.INSTALL_PIP_PACKAGE_ERROR_MESSAGE.format(package_name="kubernetes")
        ) from exc
    from kubernetes import client as k8s_client


def get_resource_config(runner_config: RunnerConfig) -> dict:
    """Get resource configuration for runner.

    :param runner_config (RunnerConfig): Runner configuration.
    :return (dict): Resource configuration.
    """
    accelerators = [
        {
            "kind": "NvidiaGpu",
            "id": gpu.device_uuid,
            "name": gpu.device_name,
            "driverVersion": gpu.device_driver_version,
            "totalVramBytes": gpu.device_vram_bytes,
            "computeCapability": gpu.device_compute_capability,
            **({"cudaCores": 1024} if gpu.device_cuda_cores is not None else {}),
        }
        for gpu in runner_config.gpus
    ]

    return {
        "id": runner_config.id,
        "metadata": {"attributes": {"displayName": runner_config.name}},
        "spec": {
            "suspended": False,
            "workers": [
                {
                    "kind": "KubernetesNode",
                    "id": str(uuid.uuid4()),
                    "resources": {
                        "cpu": runner_config.system_specs.num_cores,
                        "memoryBytes": runner_config.system_specs.ram_total,
                        "accelerators": accelerators,
                    },
                }
            ],
        },
    }


def get_role(namespace: str) -> k8s_client.V1Role:
    """Get role for runner.

    :param namespace (str): Role namespace.
    :return (k8s_client.V1Role): Role.
    """
    return k8s_client.V1Role(
        kind="Role",
        api_version="rbac.authorization.k8s.io/v1",
        metadata=k8s_client.V1ObjectMeta(name="python-job-role", namespace=namespace),
        rules=[
            k8s_client.V1PolicyRule(
                api_groups=["batch", "extensions"],
                resources=["jobs"],
                verbs=[
                    "get",
                    "list",
                    "watch",
                    "create",
                    "update",
                    "patch",
                    "delete",
                ],
            ),
            k8s_client.V1PolicyRule(
                api_groups=[""], resources=["pods"], verbs=["get", "list", "watch"]
            ),
            k8s_client.V1PolicyRule(
                api_groups=[""], resources=["pods/log"], verbs=["get", "list", "watch"]
            ),
        ],
    )


def get_role_binding(namespace: str) -> k8s_client.V1RoleBinding:
    """Get role binding for runner.

    :param namespace (str): Role binding namespace.
    :return (k8s_client.V1RoleBinding): Role binding.
    """
    return k8s_client.V1RoleBinding(
        api_version="rbac.authorization.k8s.io/v1",
        kind="RoleBinding",
        metadata=k8s_client.V1ObjectMeta(
            name="python-job-rolebinding", namespace=namespace
        ),
        subjects=[
            k8s_client.RbacV1Subject(
                kind="ServiceAccount", name="python-job-sa", namespace=namespace
            )
        ],
        role_ref=k8s_client.V1RoleRef(
            kind="Role", name="python-job-role", api_group="rbac.authorization.k8s.io"
        ),
    )


def get_deployment_spec(
    runner_config: RunnerConfig, logs_folder: Path
) -> k8s_client.V1Deployment:
    """Get deployment spec for runner.

    :param namespace (str): Deployment namespace.
    :param logs_folder (Path): Logs folder.
    :return (k8s_client.V1Deployment): Deployment specification.
    """
    return k8s_client.V1DeploymentSpec(
        replicas=1,
        selector=k8s_client.V1LabelSelector(
            match_labels={"app": RUNNER_DEPLOYMENT_NAME}
        ),
        template=k8s_client.V1PodTemplateSpec(
            metadata=k8s_client.V1ObjectMeta(
                name=RUNNER_DEPLOYMENT_NAME,
                namespace=runner_config.kubectl_namespace,
                labels={"app": RUNNER_DEPLOYMENT_NAME},
            ),
            spec=k8s_client.V1PodSpec(
                runtime_class_name="nvidia",
                service_account_name="python-job-sa",
                image_pull_secrets=[
                    k8s_client.V1LocalObjectReference(name="datature-regcred")
                ],
                containers=[
                    k8s_client.V1Container(
                        name=RUNNER_DEPLOYMENT_NAME,
                        image=os.getenv(
                            "DATATURE_RUNNER_DOCKER_REGISTRY",
                            runner_config.runner_metadata.runner_registry,
                        )
                        + f"/{runner_config.runner_metadata.runner_name}"
                        + f":{runner_config.runner_metadata.runner_version}",
                        image_pull_policy="Always",
                        env=[
                            k8s_client.V1EnvVar(
                                name="WORKSPACE_ID",
                                value_from=k8s_client.V1EnvVarSource(
                                    config_map_key_ref=k8s_client.V1ConfigMapKeySelector(
                                        name=RUNNER_DEPLOYMENT_CONFIG_NAME,
                                        key="workspace_id",
                                    )
                                ),
                            ),
                            k8s_client.V1EnvVar(
                                name="SECRET_KEY",
                                value_from=k8s_client.V1EnvVarSource(
                                    config_map_key_ref=k8s_client.V1ConfigMapKeySelector(
                                        name=RUNNER_DEPLOYMENT_CONFIG_NAME,
                                        key="secret_key",
                                    ),
                                ),
                            ),
                            k8s_client.V1EnvVar(
                                name="RUNNER_ID",
                                value_from=k8s_client.V1EnvVarSource(
                                    config_map_key_ref=k8s_client.V1ConfigMapKeySelector(
                                        name=RUNNER_DEPLOYMENT_CONFIG_NAME,
                                        key="runner_id",
                                    ),
                                ),
                            ),
                            k8s_client.V1EnvVar(
                                name="LOGS_DIRECTORY",
                                value_from=k8s_client.V1EnvVarSource(
                                    config_map_key_ref=k8s_client.V1ConfigMapKeySelector(
                                        name=RUNNER_DEPLOYMENT_CONFIG_NAME,
                                        key="logs_directory",
                                    )
                                ),
                            ),
                            k8s_client.V1EnvVar(
                                name="DOCKER_REGISTRY",
                                value_from=k8s_client.V1EnvVarSource(
                                    config_map_key_ref=k8s_client.V1ConfigMapKeySelector(
                                        name=RUNNER_DEPLOYMENT_CONFIG_NAME,
                                        key="docker_registry",
                                    )
                                ),
                            ),
                            k8s_client.V1EnvVar(
                                name="API_URL",
                                value_from=k8s_client.V1EnvVarSource(
                                    config_map_key_ref=k8s_client.V1ConfigMapKeySelector(
                                        name=RUNNER_DEPLOYMENT_CONFIG_NAME,
                                        key="api_url",
                                    )
                                ),
                            ),
                        ],
                        volume_mounts=[
                            k8s_client.V1VolumeMount(
                                name="logs",
                                mount_path="/logs",
                            )
                        ],
                    )
                ],
                volumes=[
                    k8s_client.V1Volume(
                        name="logs",
                        host_path=k8s_client.V1HostPathVolumeSource(
                            path=str(logs_folder)
                        ),
                    )
                ],
            ),
        ),
    )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   __init__.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   TensorRT init module
"""

from .tensorrt import TensorRTConverter

# Expose certain elements at package level
__all__ = ["TensorRTConverter"]



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   operation_fixture.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Operation Test Data
"""

pending_operation_response = {
    "id": "op_c26ea31d-599c-4a03-8b0d-b2bc9995fa8a",
    "object": "operation",
    "kind": "nexus.annotations.export",
    "status": {
        "overview": "Running",
        "message": "Operation running",
        "progress": {
            "unit": "asset",
            "withStatus": {
                "Queued": 10,
                "Running": 0,
                "Finished": 436,
                "Errored": 0,
                "Cancelled": 0,
            },
        },
    },
    "createDate": 1701755214003,
    "updateDate": 1701755215476,
}

finished_operation_response = {
    "id": "op_c26ea31d-599c-4a03-8b0d-b2bc9995fa8a",
    "object": "operation",
    "kind": "nexus.annotations.export",
    "status": {
        "overview": "Finished",
        "message": "Operation running",
        "payload": "",
        "progress": {
            "unit": "asset",
            "withStatus": {
                "Queued": 0,
                "Running": 0,
                "Finished": 436,
                "Errored": 0,
                "Cancelled": 0,
            },
        },
    },
    "createDate": 1701755214003,
    "updateDate": 1701755215476,
}

errored_operation_response = {
    "id": "op_c26ea31d-599c-4a03-8b0d-b2bc9995fa8a",
    "object": "operation",
    "kind": "nexus.annotations.export",
    "status": {
        "overview": "Errored",
        "message": "Operation running",
        "payload": "",
        "progress": {
            "unit": "asset",
            "withStatus": {
                "Queued": 0,
                "Running": 0,
                "Finished": 436,
                "Errored": 0,
                "Cancelled": 0,
            },
        },
    },
    "createDate": 1701755214003,
    "updateDate": 1701755215476,
}

annotation_operation_response = {
    "opId": "op_8e794c4d-47b4-460c-bcd1-173ab5d109d7",
    "status": "Finished",
    "download": {
        "method": "GET",
        "url": "https://nexus.datature.io/projects",
        "expiryDate": 1706183829833,
    },
}

annotation_export_error_response = {
    "opId": "op_8e794c4d-47b4-460c-bcd1-173ab5d109d7",
    "status": "Errored",
}

annotation_import_finished_response = {
    "id": "annotsess_de40cabf-8275-4844-9d29-1ae3f7af5ca8",
    "object": "annotation_import_session",
    "projectId": "proj_cd067221d5a6e4007ccbb4afb5966535",
    "status": {
        "overview": "Finished",
        "message": "Annotation Imported successfully.",
        "updateDate": 1705465365932,
        "files": {
            "totalSizeBytes": 63368,
            "pageCount": 1,
            "withStatus": {"Processing": 0, "Processed": 1, "FailedProcess": 0},
        },
        "annotations": {"withStatus": {"Processed": 1549, "Committed": 1549}},
    },
    "expiryDate": 1705468620676,
    "createDate": 1705465320781,
    "updateDate": 1705465365945,
}

annotation_import_errored_response = {
    "id": "annotsess_a2c771e8-6b85-47a2-b844-d25ee18e6ae3",
    "object": "annotation_import_session",
    "projectId": "proj_cd067221d5a6e4007ccbb4afb5966535",
    "status": {
        "overview": "Errored",
        "message": "AnnotationsCommitted errored, please contract support.",
        "updateDate": 1701772558131,
        "files": {
            "totalSizeBytes": 0,
            "pageCount": 1,
            "withStatus": {"Processing": 0, "Processed": 0, "FailedProcess": 0},
        },
        "annotations": {"withStatus": {"Processed": 0, "Committed": 0}},
    },
    "expiryDate": 1701768467656,
    "createDate": 1701765173451,
    "updateDate": 1701772558150,
}

annotation_import_running_response = {
    "id": "annotsess_a2c771e8-6b85-47a2-b844-d25ee18e6ae3",
    "object": "annotation_import_session",
    "projectId": "proj_cd067221d5a6e4007ccbb4afb5966535",
    "status": {
        "overview": "Running",
        "message": "Running.",
        "updateDate": 1701772558131,
        "files": {
            "totalSizeBytes": 0,
            "pageCount": 1,
            "withStatus": {"Processing": 0, "Processed": 0, "FailedProcess": 0},
        },
        "annotations": {"withStatus": {"Processed": 0, "Committed": 0}},
    },
    "expiryDate": 1701768467656,
    "createDate": 1701765173451,
    "updateDate": 1701772558150,
}



# !/usr/env/bin python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   utils.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Custom model utility functions.
"""
# pylint: disable=R0914

import yaml

from .types import (
    ConfigRequiredFields,
    InputFormatOptions,
    InputRequiredFields,
    OutputFormatOptions,
    OutputRequiredFields,
    TaskTypes,
)


def verify_graph_module(graph_module_path: str):
    """Verify the PyTorch GraphModule file.

    :param graph_module_path: Path to the PyTorch GraphModule file.

    :raises ValueError: If the PyTorch GraphModule file is not saved
        using torch.save() or the file is corrupted.
    """
    local_header_magic_number = [b"P", b"K", b"\x03", b"\x04"]
    read_bytes = []

    with open(graph_module_path, "rb") as graph_module_reader:
        for _ in range(4):
            byte = graph_module_reader.read(1)

            if byte == b"":
                break

            read_bytes.append(byte)

    if read_bytes != local_header_magic_number:
        raise ValueError(
            "PyTorch GraphModule does not seem to be saved using "
            "torch.save() or the file is corrupted. "
            "Please save the model using torch.save() and try again."
        )


def verify_yaml_config(yaml_file: str):
    """Verify the YAML configuration file.

    :param yaml_file: Path to the YAML configuration file.
    """
    with open(yaml_file, "r", encoding="utf-8") as yaml_reader:
        yaml_content = yaml.safe_load(yaml_reader)

    assert yaml_content, "YAML file is empty. Please provide a valid YAML file."
    for config_field in ConfigRequiredFields:
        assert config_field.value in yaml_content, (
            f"Config field '{config_field.value}' is missing from the YAML file. "
            "Please provide the missing field."
        )

    task = yaml_content["task"]
    assert task in TaskTypes, f"Task type '{task}' needs to be one of {TaskTypes}"

    inputs = yaml_content["inputs"]
    assert inputs, (
        "No input fields found in the YAML file. "
        "Please provide at least one input field."
    )

    input_required_fields = InputRequiredFields[task.upper()].value
    for input_field in input_required_fields:
        assert inputs.get(input_field), (
            f"Input field '{input_field}' is missing from the YAML file. "
            "Please provide the missing field."
        )

    for key, value in inputs.items():
        assert isinstance(
            value["name"], str
        ), f"Input name '{value['name']}' must be a string"
        assert isinstance(
            value["key"], str
        ), f"Input key '{value['key']} must be a string"
        assert isinstance(
            value["format"], list
        ), f"Input format '{value['format']}' must be a list"
        assert value["format"] in InputFormatOptions, (
            f"Input format '{value['format']}' "
            f"needs to be one of {InputFormatOptions}"
        )
        assert isinstance(
            value["shape"], list
        ), f"Input shape '{value['shape']}' must be a list"
        assert len(value["shape"]) == len(value["format"]), (
            f"Input shape {len(value['shape'])} and input format "
            f"{len(value['format'])} must have the same number of elements"
        )

    outputs = yaml_content["outputs"]
    assert outputs, (
        "No output fields found in the YAML file. "
        "Please provide at least one output field."
    )

    output_required_fields = OutputRequiredFields[task.upper()].value
    for output_field in output_required_fields:
        assert outputs.get(output_field), (
            f"Output field '{output_field}' is missing from the YAML file. "
            "Please provide the missing field."
        )

    output_format_options = OutputFormatOptions[task.upper()].value
    for key, value in outputs.items():
        assert isinstance(
            value["name"], str
        ), f"Output name '{value['name']}' must be a string"
        assert isinstance(
            value["key"], str
        ), f"Output key '{value['key']} must be a string"

        if key in output_required_fields:
            assert value["format"] in output_format_options[key.upper()].value, (
                f"Output format '{value['format']}' "
                f"needs to be one of {output_format_options[key.upper()].value}"
            )

        assert isinstance(
            value["shape"], list
        ), f"Output shape '{value['shape']}' must be a list"
        assert len(value["shape"]) == len(value["format"]), (
            f"Output shape {len(value['shape'])} and format "
            f"{len(value['format'])} must have the same number of elements"
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   consts.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   CLI Constants.
"""

from enum import Enum, unique

from datature.nexus.api.types import add_mapping

INQUIRER_CHOICES_MAX_VISIBLE_ENTRIES = 10

DEFAULT_INSTANCES = [
    "instance_x1cpu-standard",
    "instance_t4-standard-1g",
    "instance_l4-standard-1g",
]


@add_mapping
@unique
class ResourceType(Enum):
    """Resource type enumeration."""

    GPU = "GPU"
    CPU = "CPU"

    __MAPPING__ = {
        "GPU": GPU,
        "CPU": CPU,
    }



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   Deploy.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Deploy API
"""

from typing import Union

from datature.nexus import models
from datature.nexus.api.types import DeploymentMetadata
from datature.nexus.client_context import RestContext


class Deployment(RestContext):
    """Datature Deploy API Resource."""

    def list(self) -> models.Deployments:
        """Lists all deployments in a project.

        :return: A list of msgspec structs containing
            deployment metadata with the following structure:

            .. code-block:: python

                [Deployment(
                    id='deploy_1c144673-c757-40b4-8eeb-d400f0b9b2f9',
                    name='My API',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    artifact_id='artifact_65a7678a91afa7d22455c5ba',
                    version_tag='v1.0.0',
                    history_versions=[
                        DeploymentHistoryVersion(
                            version_tag='v1.0.0',
                            artifact_id='artifact_65a7678a91afa7d22455c5ba',
                            update_date=1705475663570
                        )
                    ],
                    resources=DeploymentResources(
                        cpu=4,
                        ram=8192,
                        GPU_T4=None
                    ),
                    scaling=DeploymentScaling(
                        replicas=1,
                        mode='FixedReplicaCount'
                    ),
                    status=DeploymentStatus(
                        overview='Available',
                        message='Created service successfully',
                        update_data=1705475727032
                    ),
                    create_date=1705475663570,
                    update_date=1705475727051,
                    url='https://asia-inference.nip.io/1c144673-c757-40b4-8eeb-d400f0b9b2f9'
                )]

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")
                        project.deployments.list()
        """
        return self.requester.GET(
            f"/projects/{self.project_id}/deployments", response_type=models.Deployments
        )

    def get(self, deploy_id: str) -> models.Deployment:
        """Retrieves a specific deployment using the deployment ID.

        :param deploy_id: The ID of the deployment as a string.
        :return: A msgspec struct containing the
            specific deployment metadata with the following structure:

            .. code-block:: python

                Deployment(
                    id='deploy_1c144673-c757-40b4-8eeb-d400f0b9b2f9',
                    name='My API',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    artifact_id='artifact_65a7678a91afa7d22455c5ba',
                    version_tag='v1.0.0',
                    history_versions=[
                        DeploymentHistoryVersion(
                            version_tag='v1.0.0',
                            artifact_id='artifact_65a7678a91afa7d22455c5ba',
                            update_date=1705475663570
                        )
                    ],
                    resources=DeploymentResources(
                        cpu=4,
                        ram=8192,
                        GPU_T4=None
                    ),
                    scaling=DeploymentScaling(
                        replicas=1,
                        mode='FixedReplicaCount'
                    ),
                    status=DeploymentStatus(
                        overview='Available',
                        message='Created service successfully',
                        update_data=1705475727032
                    ),
                    create_date=1705475663570,
                    update_date=1705475727051,
                    url='https://asia-inference.nip.io/1c144673-c757-40b4-8eeb-d400f0b9b2f9'
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")
                project.deployments.get("deploy_1c144673-c757-40b4-8eeb-d400f0b9b2f9")
        """
        assert isinstance(deploy_id, str)
        return self.requester.GET(
            f"/projects/{self.project_id}/deployments/{deploy_id}",
            response_type=models.Deployment,
        )

    def delete(self, deploy_id: str) -> models.DeleteResponse:
        """Deletes a specific deployment from the project.

        :param deploy_id: The id of the deployment.
        :return: A msgspec struct containing the deleted
            deployment ID and the deletion status with the following structure:

            .. code-block:: python

                DeleteResponse(deleted=True, id="deploy_30922d5e-b2f6-43dc-b7b4-e29e2c30fb45")

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")
                project.deployments.delete("deploy_30922d5e-b2f6-43dc-b7b4-e29e2c30fb45")
        """
        assert isinstance(deploy_id, str)
        return self.requester.DELETE(
            f"/projects/{self.project_id}/deployments/{deploy_id}",
            response_type=models.DeleteResponse,
        )

    def create(self, deployment: Union[DeploymentMetadata, dict]) -> models.Deployment:
        """Creates a deployment for a specific model using the model ID.

        :param deployment: The configuration metadata of the deployment.
        :return: A msgspec struct containing the specific
            deployment metadata with the following structure:

            .. code-block:: python

                Deployment(
                    id='deploy_1c144673-c757-40b4-8eeb-d400f0b9b2f9',
                    name='My API',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    artifact_id='artifact_65a7678a91afa7d22455c5ba',
                    version_tag='v1.0.0',
                    history_versions=[
                        DeploymentHistoryVersion(
                            version_tag='v1.0.0',
                            artifact_id='artifact_65a7678a91afa7d22455c5ba',
                            update_date=1705475663570
                        )
                    ],
                    resources=DeploymentResources(
                        cpu=4,
                        ram=8192,
                        GPU_T4=None
                    ),
                    scaling=DeploymentScaling(
                        replicas=1,
                        mode='FixedReplicaCount'
                    ),
                    status=DeploymentStatus(
                        overview='Creating',
                        message='Creating service.',
                        update_data=1705475727032
                    ),
                    create_date=1705475663570,
                    update_date=1705475727051,
                    url=None
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.deployments.create({
                    "name": "My First API",
                    "artifact_id": "artifact_63fd950a64845427a706d57c",
                    "version_tag": "car.staging.v3",
                })
        """
        assert isinstance(deployment, (DeploymentMetadata, dict))

        if isinstance(deployment, dict):
            deployment = DeploymentMetadata(**deployment)

        if deployment.instance_id:
            available_instances = self.requester.GET(
                f"/projects/{self.project_id}/artifacts/{deployment.artifact_id}"
                f"/availableInstances/regions/{deployment.region if deployment.region else '*'}",
                response_type=models.AvailableInstances,
            )
            available_instance_ids = [
                instance.id for instance in available_instances.instances
            ]
            if deployment.instance_id not in available_instance_ids:
                raise ValueError(
                    f"Instance {deployment.instance_id} is not available. Available instances: "
                    f"{available_instance_ids}"
                )

        return self.requester.POST(
            f"/projects/{self.project_id}/deployments",
            request_body=deployment.to_json(),
            response_type=models.Deployment,
        )

    def update(
        self, deploy_id: str, deployment: Union[DeploymentMetadata, dict]
    ) -> models.Deployment:
        """Creates a deployment for a specific model using the model ID.

        :param deploy_id: The ID of the deployment as a string.
        :param deployment: The configuration metadata of the deployment.
        :return: A msgspec struct containing the
            specific deployment metadata with the following structure:

            .. code-block:: python

                Deployment(
                    id='deploy_1c144673-c757-40b4-8eeb-d400f0b9b2f9',
                    name='My First API v2',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    artifact_id='artifact_65a7678a91afa7d22455c5ba',
                    version_tag='v1.0.0',
                    history_versions=[DeploymentHistoryVersion(
                        version_tag='v1.0.0',
                        artifact_id='artifact_65a7678a91afa7d22455c5ba',
                        update_date=1705475663570
                    )],
                    resources=DeploymentResources(
                        cpu=4,
                        ram=8192,
                        GPU_T4=1
                    ),
                    scaling=DeploymentScaling(
                        replicas=1,
                        mode='FixedReplicaCount'
                    ),
                    status=DeploymentStatus(
                        overview='Creating',
                        message='Creating service',
                        update_data=1705475679244
                    ),
                    create_date=1705475663570,
                    update_date=1705477694540,
                    url=None
                )

        :example:

            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                datature.deployments.update(
                    "deploy_30922d5e-b2f6-43dc-b7b4-e29e2c30fb45",
                    {
                        "name": "My First API v2",
                        "options": {
                            "evaluation_threshold": 0.8
                        }
                        "resources":{
                            "GPU_T4":1
                        }
                })
        """
        assert isinstance(deployment, (DeploymentMetadata, dict))

        if isinstance(deployment, dict):
            deployment = DeploymentMetadata(**deployment)

        return self.requester.PATCH(
            f"/projects/{self.project_id}/deployments/{deploy_id}",
            request_body=deployment.to_json(),
            response_type=models.Deployment,
        )

    def create_version(
        self, deploy_id: str, version_tag: str, artifact_id: str
    ) -> models.Deployment:
        """Updates a deployment version for a specific artifact using the artifact ID.

        :param deploy_id: The ID of the deployment as a string.
        :param version_tag: The new version tag name of the deployment.
        :param artifact_id: The ID of the artifact as a string.
        :return: A msgspec struct containing the specific
            deployment metadata with the following structure:

            .. code-block:: python

                Deployment(
                    id='deploy_1c144673-c757-40b4-8eeb-d400f0b9b2f9',
                    name='My API',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    artifact_id='artifact_65a75d4891afa7d22455c54d',
                    version_tag='v2.0.0',
                    history_versions=[
                        DeploymentHistoryVersion(
                            version_tag='v1.0.0',
                            artifact_id='artifact_65a7678a91afa7d22455c5ba',
                            update_date=1705475663570
                        ),
                        DeploymentHistoryVersion(
                            version_tag='v2.0.0',
                            artifact_id='artifact_65a75d4891afa7d22455c54d',
                            update_date=1705477973476
                        )
                    ],
                    resources=DeploymentResources(
                        cpu=4,
                        ram=8192,
                        GPU_T4=1
                    ),
                    scaling=DeploymentScaling(
                        replicas=1,
                        mode='FixedReplicaCount'
                    ),
                    status=DeploymentStatus(
                        overview='Creating',
                        message='Creating service',
                        update_data=1705475679244
                    ),
                    create_date=1705475663570,
                    update_date=1705477694540,
                    url=None
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.deployments.create_version(
                    "deploy_30922d5e-b2f6-43dc-b7b4-e29e2c30fb45",
                    "v2.0.0",
                    "artifact_63fd950a64845427a706d57d"
                )
        """
        assert isinstance(deploy_id, str)
        assert isinstance(version_tag, str)
        assert isinstance(artifact_id, str)

        return self.requester.POST(
            f"/projects/{self.project_id}/deployments/{deploy_id}/versions",
            request_body={
                "versionTag": version_tag,
                "artifactId": artifact_id,
            },
            response_type=models.Deployment,
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   project.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Project API
"""
# pylint: disable=R0902

from abc import ABC
from collections import ChainMap
from inspect import isclass
from typing import Union

from datature.nexus import models
from datature.nexus.api.annotation.annotation import Annotation
from datature.nexus.api.artifact import Artifact
from datature.nexus.api.asset.asset import Asset
from datature.nexus.api.batch.batch import Batch
from datature.nexus.api.deployment import Deployment
from datature.nexus.api.ontology import Ontology
from datature.nexus.api.operation import Operation
from datature.nexus.api.run import Run
from datature.nexus.api.tag import Tag
from datature.nexus.api.types import ProjectMetadata
from datature.nexus.api.workflow import Workflow
from datature.nexus.client_context import ClientContext, RestContext


class Project(ABC):
    """Datature Project API Resource."""

    _context: ClientContext
    operations: Operation
    assets: Asset
    annotations: Annotation
    tags: Tag
    workflows: Workflow
    artifacts: Artifact
    runs: Run
    deployments: Deployment
    ontologies: Ontology
    batch: Batch

    def __init__(self, client_context: ClientContext):
        """Initialize the Project API Resource."""
        self.requester = client_context.requester
        self.project_id = client_context.project_id
        self._context = client_context

        # Init sub resources
        self._auto_init_resources()

    def _get_all_annotations(self):
        """Get annotations for this class and all superclasses."""
        # This expression correctly handles cases where multiple superclasses
        # provide different annotations for the same attribute name.
        # See https://stackoverflow.com/a/72037059.
        return ChainMap(
            *(
                c.__annotations__
                for c in self.__class__.__mro__
                if "__annotations__" in c.__dict__
            )
        )

    def _auto_init_resources(self):
        for name, type_annot in self._get_all_annotations().items():
            # Don't overwrite existing fields
            if hasattr(self, name):
                continue

            # Only initialize if the type is a class and a subclass of RestContext
            if not isclass(type_annot) or not issubclass(type_annot, RestContext):
                continue

            setattr(self, name, type_annot(self._context))

    def get_info(self) -> models.Project:
        """Retrieves project information.

        :return: A msgspec struct containing the projects with the following structure:

            .. code-block:: python

                Project(
                    id='proj_9004a21df7b040ace4674c4879603fe8',
                    name='keypoints',
                    workspace_id='ws_1c8aab980f174b0296c7e35e88665b13',
                    type='ObjectDetection',
                    create_date=1701927649302,
                    localization='MULTI',
                    tags=['cat faces'],
                    groups=['main', 'cats'],
                    statistic=Statistic(
                        tags_count=[TagsCountItem(name='cat faces', count=0)],
                        total_assets=28,
                        annotated_assets=0,
                        total_annotations=0
                    )
                )


        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")
                project.get_info()
        """
        project = self._context.requester.GET(
            f"/projects/{self.project_id}", response_type=models.Project
        )

        return project

    def update(self, project: Union[ProjectMetadata, dict]) -> models.Project:
        """Updates the project with project meta fields.

        :param project: The new metadata of the project to be updated.
        :return: A msgspec struct containing the projects with the following structure:

            .. code-block:: python

                Project(
                    id='proj_9004a21df7b040ace4674c4879603fe8',
                    name='My Cool Project',
                    workspace_id='ws_1c8aab980f174b0296c7e35e88665b13',
                    type='ObjectDetection',
                    create_date=1701927649302,
                    localization='MULTI',
                    tags=['cat faces'],
                    groups=['main', 'cats'],
                    statistic=Statistic(
                        tags_count=[TagsCountItem(name='cat faces', count=0)],
                        total_assets=28,
                        annotated_assets=0,
                        total_annotations=0
                    )
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client, ApiTypes

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.update({"name":"My Cool Project"})
                // Or
                project.update(ApiTypes.ProjectMetadata(name="My Cool Project"))
        """
        assert isinstance(project, (ProjectMetadata, dict))

        if isinstance(project, dict):
            project = ProjectMetadata(**project)

        return self.requester.PATCH(
            f"/projects/{self.project_id}",
            request_body=project.to_json(),
            response_type=models.Project,
        )

    def list_insights(self) -> models.ProjectInsights:
        """Retrieves project insight and metrics of the completed training runs.

        :return: A msgspec struct containing
            the project insights metadata with the following structure:

            .. code-block:: python

                [ProjectInsight(
                    flow_title='Test workflow',
                    run_id='run_4a5d406d-464d-470c-bd7d-e92456621ad3',
                    dataset=InsightDataset(
                        data_type='Rectangle',
                        num_classes=1,
                        average_annotations=5.19,
                        total_assets=500,
                        settings=DatasetSettings(
                            split_ratio=0.3,
                            shuffle=True,
                            seed=0,
                            using_sliding_window=False
                        )
                    ),
                    model=InsightModel(
                        name='fasterrcnn-inceptionv2-1024x1024',
                        batch_size=2,
                        training_steps=5000,
                        max_detection_per_class=100,
                        solver='momentum',
                        learning_rate=0.04,
                        momentum=0.9
                    ),
                    checkpoint=RunCheckpoint(
                        strategy='STRAT_ALWAYS_SAVE_LATEST',
                        evaluation_interval=250,
                        metric=None
                    ),
                    artifact=InsightArtifact(
                        id='artifact_65ae274540259e2a07533532',
                        is_training=False,
                        step=5000,
                        metric=ArtifactMetric(
                            total_loss=0.32356,
                            classification_loss=0.012036,
                            localization_loss=0.010706,
                            regularization_loss=0.0
                        )
                    ),
                    create_date=1705912133684
                )]

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")
                project.list_insights()
        """
        return self.requester.GET(
            f"/projects/{self.project_id}/insights",
            response_type=models.ProjectInsights,
        )

    def list_users(self) -> models.ProjectUsers:
        """Retrieves all users in the project.
            This includes Project Owners, Collaborators and Datature Experts.

        :return: A list of msgspec struct containing
            the project user metadata with the following structure:

            .. code-block:: python

                [ProjectUser(
                    id='user_6323fea23e292439f31c58cd',
                    access_type='Owner',
                    email='raighne@datature.io',
                    nickname='raighne',
                    picture='https://s.gravatar.com/avatar/avatars%2Fra.png'
                )]

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")
                project.list_users()
        """
        return self.requester.GET(
            f"/projects/{self.project_id}/users", response_type=models.ProjectUsers
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_tag.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Project API Test Cases.
"""

import unittest
from unittest.mock import ANY, MagicMock

from datature.nexus.api.tag import Tag
from datature.nexus.client_context import ClientContext

# pylint: disable=W0212


class TestTag(unittest.TestCase):
    """Datature Tag API Resource Test Cases."""

    def __init__(self, *args, **kwargs):
        """init global variables."""
        unittest.TestCase.__init__(self, *args, **kwargs)

        context = ClientContext("secret_key", endpoint="http://localhost:8080")
        context.project_id = "project_id"
        context.requester._request = MagicMock()

        self.tags = Tag(context)

    def test_list(self):
        """Test list tags."""
        self.tags.list()

        self.tags._context.requester._request.assert_called_once_with(
            "GET", "/projects/project_id/tags", ANY, query=None, request_params=None
        )

    def test_create(self):
        """Test create a tag."""
        self.tags.create({"name": "New Tag Name"})

        self.tags._context.requester._request.assert_called_once_with(
            "POST",
            "/projects/project_id/tags",
            ANY,
            request_body={"name": "New Tag Name"},
            query=None,
            request_params=None,
        )

    def test_create_with_metadata(self):
        """Test create a tag with metadata."""
        self.tags.create({"name": "New Tag Name", "color": "#000000"})

        self.tags._context.requester._request.assert_called_once_with(
            "POST",
            "/projects/project_id/tags",
            ANY,
            request_body={"name": "New Tag Name", "color": "#000000"},
            query=None,
            request_params=None,
        )

    def test_update(self):
        """Test update a tag."""
        self.tags.update(1, {"name": "New Tag Name"})

        self.tags._context.requester._request.assert_called_once_with(
            "PATCH",
            "/projects/project_id/tags/1",
            ANY,
            request_body={"name": "New Tag Name"},
            query=None,
            request_params=None,
        )

    def test_update_with_metadata(self):
        """Test update a tag with metadata."""
        self.tags.update(1, {"name": "New Tag Name", "color": "#000000"})

        self.tags._context.requester._request.assert_called_once_with(
            "PATCH",
            "/projects/project_id/tags/1",
            ANY,
            request_body={"name": "New Tag Name", "color": "#000000"},
            query=None,
            request_params=None,
        )

    def test_delete(self):
        """Test delete a tag."""
        self.tags.delete(1)

        self.tags._context.requester._request.assert_called_once_with(
            "DELETE", "/projects/project_id/tags/1", ANY
        )

    def test_merge(self):
        """Test merge a tag."""
        self.tags.merge(1, 0)

        self.tags._context.requester._request.assert_called_once_with(
            "POST",
            "/projects/project_id/tags/1-0:merge",
            ANY,
            query=None,
            request_body=None,
            request_params=None,
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   webhook.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   CLI Webhook Functions.
"""

import logging
import sys
from datetime import datetime
from pathlib import Path

from colorama import Fore
from InquirerPy import inquirer
from InquirerPy.base.control import Choice
from InquirerPy.validator import EmptyInputValidator
from wcwidth import wcswidth

import datature.nexus.cli.batch.functions as batch_functions
from datature.nexus import error
from datature.nexus.cli.functions import get_default_datature_client
from datature.nexus.utils import utils

logging.basicConfig(level=logging.CRITICAL)


def create_webhook_secret(prompt_for_save: bool = True) -> str:
    """
    Creates a Webhook secret key and either prints it to the console or saves it to a file.

    :param: prompt_for_save: Whether to prompt the user to save the secret key to a file.
    :return: The generated Webhook secret key.
    """
    webhook_secret = utils.generate_webhook_secret()

    if prompt_for_save:
        webhook_secret_filepath = (
            inquirer.filepath(
                "Enter the path to the file where the generated Webhook secret key will be saved:",
                mandatory=False,
                long_instruction="Leave blank to print to console",
            )
            .execute()
            .strip()
        )

        if webhook_secret_filepath:
            try:
                webhook_secret_filepath = (
                    Path(webhook_secret_filepath).expanduser().resolve()
                )
                if not webhook_secret_filepath.parent.exists():
                    webhook_secret_filepath.parent.mkdir(parents=True, exist_ok=True)
                webhook_secret_filepath.write_text(webhook_secret, encoding="utf-8")
            except (FileNotFoundError, PermissionError) as exc:
                print(
                    f"{Fore.CYAN}●{Fore.RESET} Error creating and writing to file: {exc}. "
                    "Please ensure that the file path is correct and that "
                    "you have the necessary write permissions."
                )
                sys.exit(1)

    print(f"{Fore.GREEN}✓{Fore.RESET} Webhook secret key generated successfully:")
    print(f"\n  {webhook_secret}\n")
    return webhook_secret


def handle_webhook_secret() -> str:
    """
    Handles the generation of a Webhook secret key.

    :return: The Webhook secret key.
    """
    webhook_secret_generation_method = inquirer.select(
        message="Select how you would like Datature to sign Webhook requests:",
        choices=[
            Choice(name="Generate a new Webhook secret key", value="new"),
            Choice(
                name="Load an existing Webhook secret key from a file", value="file"
            ),
            Choice(name="Enter an existing Webhook secret key", value="input"),
        ],
        border=True,
    ).execute()

    if webhook_secret_generation_method == "new":
        webhook_secret = create_webhook_secret()
    elif webhook_secret_generation_method == "file":
        webhook_secret_filepath = (
            inquirer.filepath(
                "Enter the path to the file containing the secret key:",
                validate=utils.PathAndEmptyInputValidator(),
            )
            .execute()
            .strip()
        )

        try:
            webhook_secret = (
                Path(webhook_secret_filepath)
                .expanduser()
                .resolve()
                .read_text(encoding="utf-8")
                .strip()
            )
        except (FileNotFoundError, PermissionError) as exc:
            print(
                f"{Fore.CYAN}●{Fore.RESET} Error reading file: {exc}. "
                "Please ensure the path is correct and you have the necessary read permissions."
            )
            sys.exit(1)
    else:
        webhook_secret = (
            inquirer.secret("Enter the secret key:", validate=EmptyInputValidator())
            .execute()
            .strip()
        )
    return webhook_secret


def create_webhook(webhook_type: str = "custom") -> str:
    """
    Creates a new Webhook.

    :param webhook_type: The type of the Webhook to create.
    :return: The ID of the created Webhook.
    """
    sdk_project = get_default_datature_client()

    webhook_name = (
        (
            inquirer.text(
                "Enter a name for the Webhook:",
                validate=utils.NameAndEmptyInputValidator(),
            )
            .execute()
            .strip()
        )
        if webhook_type == "custom"
        else f"webhook-{datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}"
    )

    webhook_endpoint = (
        inquirer.text("Enter the Webhook endpoint:", validate=EmptyInputValidator())
        .execute()
        .strip()
    )

    max_retries = (
        int(
            inquirer.number(
                "Enter the maximum number of retries per request [0-5]:",
                default=3,
                min_allowed=0,
                max_allowed=5,
            ).execute()
        )
        if webhook_type == "custom"
        else 3
    )

    max_delay = (
        int(
            inquirer.number(
                "Enter the maximum delay between retries in seconds [0-60]:",
                default=15,
                min_allowed=0,
                max_allowed=60,
            ).execute()
        )
        * 1000
        if webhook_type == "custom"
        else 15000
    )

    webhook_secret = (
        handle_webhook_secret()
        if webhook_type == "custom"
        else create_webhook_secret(prompt_for_save=False)
    )

    try:
        webhook = sdk_project.batch.webhooks.create(
            name=webhook_name,
            webhook_spec={
                "endpoint": webhook_endpoint,
                "secret": {"contents": webhook_secret},
                "retries": {
                    "max_retries": int(max_retries),
                    "max_retry_delay": int(max_delay),
                },
            },
        )
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error creating Webhook: {exc.message}")
        sys.exit(1)

    print(
        f"{Fore.GREEN}✓{Fore.RESET} Webhook created successfully. "
        f"Run 'datature batch webhooks test {webhook.id}' to test the Webhook."
    )

    batch_functions.print_webhook_object(webhook)
    return webhook.id


def get_webhook(webhook_id: str) -> None:
    """
    Gets a Webhook by ID.

    :param webhook_id: The ID of the Webhook to get.
    """
    sdk_project = get_default_datature_client()

    if not webhook_id:
        webhook_id = batch_functions.select_entry("webhooks")

    try:
        webhook = sdk_project.batch.webhooks.get(webhook_id)
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error getting Webhook: {exc.message}")
        sys.exit(1)

    batch_functions.print_webhook_object(webhook)


def list_webhooks() -> None:
    """
    Lists all Webhooks.
    """
    sdk_project = get_default_datature_client()

    try:
        webhooks_response = sdk_project.batch.webhooks.list()
        webhooks_response.data.reverse()
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error listing Webhooks: {exc.message}")
        sys.exit(1)

    if len(webhooks_response.data) == 0:
        print(f"{Fore.CYAN}●{Fore.RESET} No Webhooks found.")
        return

    print()  # for new line
    print(f"{'ID':<45} {'NAME':<20} {'ENDPOINT':<20}")
    for webhook in webhooks_response.data:
        name_padding = max(0, 20 - (wcswidth(webhook.name) - len(webhook.name)))
        endpoint_padding = max(
            0, 20 - (wcswidth(webhook.endpoint) - len(webhook.endpoint))
        )
        print(
            f"{webhook.id:<45} {utils.truncate_text(webhook.name, 20):<{name_padding}} "
            f"{utils.truncate_text(webhook.endpoint, 20):<{endpoint_padding}}"
        )
    print()  # for new line


def delete_webhook(webhook_id: str) -> None:
    """
    Deletes a Webhook by ID.

    :param webhook_id: The ID of the Webhook to delete.
    """
    sdk_project = get_default_datature_client()

    if not webhook_id:
        webhook_id = batch_functions.select_entry("webhooks")

    try:
        sdk_project.batch.webhooks.delete(webhook_id)
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error deleting Webhook: {exc.message}")
        sys.exit(1)
    print(f"{Fore.GREEN}✓{Fore.RESET} Webhook deleted successfully.")


def update_webhook(webhook_id: str) -> None:
    """
    Updates a Webhook by ID.

    :param webhook_id: The ID of the Webhook to update.
    """
    sdk_project = get_default_datature_client()

    if not webhook_id:
        webhook_id = batch_functions.select_entry("webhooks")

    try:
        webhook = sdk_project.batch.webhooks.get(webhook_id)
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error getting Webhook: {exc.message}")
        sys.exit(1)

    webhook_endpoint = (
        inquirer.text(
            "Enter the new Webhook endpoint:",
            long_instruction=f"Leave blank to keep current: {webhook.endpoint}",
        )
        .execute()
        .strip()
    ) or webhook.endpoint

    max_retries = (
        inquirer.number(
            "Enter the maximum number of retries per request [0-5]:",
            long_instruction=f"Leave blank to keep current: {webhook.retries.max_retries}",
            default=None,
            min_allowed=0,
            max_allowed=5,
        ).execute()
        or webhook.retries.max_retries
    )

    max_delay = (
        inquirer.number(
            "Enter the maximum delay between retries in seconds [0-60]:",
            long_instruction=f"Leave blank to keep current: {int(webhook.retries.max_retry_delay / 1000)}",
            default=None,
            min_allowed=0,
            max_allowed=60,
        ).execute()
        or webhook.retries.max_retry_delay
    )

    try:
        updated_webhook = sdk_project.batch.webhooks.update(
            webhook_id=webhook_id,
            webhook_spec={
                "endpoint": webhook_endpoint,
                "retries": {
                    "max_retries": int(max_retries),
                    "max_retry_delay": int(max_delay),
                },
            },
        )
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error updating Webhook: {exc.message}")
        sys.exit(1)

    print()  # for new line
    print(f"{Fore.GREEN}✓{Fore.RESET} Webhook updated successfully.")

    batch_functions.print_webhook_object(updated_webhook)


def update_webhook_secret(webhook_id: str) -> None:
    """
    Updates the secret key of a Webhook by ID.

    :param webhook_id: The ID of the Webhook to update.
    """
    sdk_project = get_default_datature_client()

    if not webhook_id:
        webhook_id = batch_functions.select_entry("webhooks")

    try:
        sdk_project.batch.webhooks.get(webhook_id)
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error getting Webhook: {exc.message}")
        sys.exit(1)

    webhook_secret = handle_webhook_secret()
    try:
        sdk_project.batch.webhooks.update_secret(
            webhook_id=webhook_id, secret=webhook_secret
        )
    except error.Error as exc:
        print(
            f"{Fore.CYAN}●{Fore.RESET} Error updating Webhook secret key: {exc.message}"
        )
        sys.exit(1)

    print(f"{Fore.GREEN}✓{Fore.RESET} Webhook secret key updated successfully.")


def test_webhook(webhook_id: str) -> None:
    """
    Tests a Webhook by ID by sending a sample payload and checking the response.

    :param webhook_id: The ID of the Webhook to ping.
    """
    sdk_project = get_default_datature_client()

    if not webhook_id:
        webhook_id = batch_functions.select_entry("webhooks")

    try:
        webhook_response = sdk_project.batch.webhooks.test(webhook_id)
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error testing Webhook: {exc.message}")
        sys.exit(1)

    header_width = 15
    print()  # for new line
    print("=== Webhook Test Response ===")
    print(f"{'Status:':<{header_width}} {webhook_response.status}")
    print(f"{'Response Code:':<{header_width}} {webhook_response.response_code}")
    print(f"{'Latency (ms):':<{header_width}} {webhook_response.latency_ms}")
    print(f"{'Attempt Count':<{header_width}} {webhook_response.attempt_count}")
    print(f"{'Response Body:':<{header_width}} {webhook_response.body}")
    print(f"{'Reason:':<{header_width}} {webhook_response.reason}")
    print()  # for new line



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   run.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Run API
"""

from typing import Union

from datature.nexus import models
from datature.nexus.api.types import RunSetupMetadata
from datature.nexus.client_context import RestContext


class Run(RestContext):
    """Datature Run API Resource."""

    def list(self) -> models.Runs:
        """Lists all training runs regardless of status.

        :return: A list of msgspec structs containing
            the training run metadata with the following structure:

            .. code-block:: python

                [Run(
                    id='run_610ba26c-6fbb-42eb-b838-839c61b68b26',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    flow_id='flow_63bbd3bf8a78eb906f417396',
                    execution=RunExecution(
                        accelerator=RunAccelerator(name='GPU_T4', count=1),
                        checkpoint=RunCheckpoint(
                            strategy='STRAT_ALWAYS_SAVE_LATEST',
                            evaluation_interval=250,
                            metric=None
                        ),
                        limit=RunLimit(metric='LIM_NONE', value=0)
                    ),
                    features=RunFeatures(preview=True, matrix=True, using_sliding_window=False),
                    status=RunStatus(
                        overview='Creating',
                        message='Creating service.',
                        update_data=1705466384796
                    ),
                    create_date=1705466384796,
                    update_date=1705466392067,
                    log_ids=['runlog_UjYxMGJhMjZjLTZmYmItNDJlYi1iODM4LTgzOWM2MWI2OGIyNg']
                )]

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.runs.list()
        """
        return self.requester.GET(
            f"/projects/{self.project_id}/runs", response_type=models.Runs
        )

    def get(self, run_id: str) -> models.Run:
        """Retrieves a specific training run using the run ID.

        :param run_id: The ID of the training run.
        :return: A msgspec struct containing the specific
            training run metadata with the following structure:

            .. code-block:: python

                Run(
                    id='run_610ba26c-6fbb-42eb-b838-839c61b68b26',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    flow_id='flow_63bbd3bf8a78eb906f417396',
                    execution=RunExecution(
                        accelerator=RunAccelerator(name='GPU_T4', count=1),
                        checkpoint=RunCheckpoint(
                            strategy='STRAT_ALWAYS_SAVE_LATEST',
                            evaluation_interval=250,
                            metric=None
                        ),
                        limit=RunLimit(metric='LIM_NONE', value=0)
                    ),
                    features=RunFeatures(preview=True, matrix=True, using_sliding_window=False),
                    status=RunStatus(
                        overview='Creating',
                        message='Creating service.',
                        update_data=1705466384796
                    ),
                    create_date=1705466384796,
                    update_date=1705466392067,
                    log_ids=['runlog_UjYxMGJhMjZjLTZmYmItNDJlYi1iODM4LTgzOWM2MWI2OGIyNg']
                )

        :example:
            .. code-block:: python

                    from datature.nexus import Client

                    project = Client("5aa41e8ba........").get_project("proj_b705a........")

                    project.runs.get("run_610ba26c-6fbb-42eb-b838-839c61b68b26")
        """
        assert isinstance(run_id, str)

        return self.requester.GET(
            f"/projects/{self.project_id}/runs/{run_id}", response_type=models.Run
        )

    def kill(self, run_id: str) -> models.Run:
        """Kills a specific training run using the run ID.

        :param run_id: The ID of the training run.
        :return: A msgspec struct containing the
            killed training metadata with the following structure:

            .. code-block:: python

                Run(
                    id='run_e2a14cee-eacc-4335-bc95-94c3ee196b04',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    flow_id='flow_64e812a7e47592ef374cbbc2',
                    execution=RunExecution(
                        accelerator=RunAccelerator(name='GPU_L4', count=1),
                        checkpoint=RunCheckpoint(
                            strategy='STRAT_LOWEST_VALIDATION_LOSS',
                            evaluation_interval=220,
                            metric='Loss/total_loss'
                        ),
                        limit=RunLimit(metric='LIM_NONE', value=0)
                    ),
                    status=RunStatus(
                        overview='Cancelled',
                        message='Training cancelled.',
                        update_data=1700204316180
                    ),
                    create_date=1701927649302,
                    update_date=1701927649302,
                    features=RunFeatures(preview=True, matrix=True),
                    log_ids=['runlog_UmUyYTE0Y2VlLWVhY2MtNDMzNS1iYzk1LTk0YzNlZTE5NmIwNA']
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.runs.kill("run_63eb212ff0f856bf95085095")
        """
        assert isinstance(run_id, str)

        return self.requester.PATCH(
            f"/projects/{self.project_id}/runs/{run_id}",
            request_body={"status": "Cancelled"},
            response_type=models.Run,
        )

    def start(self, flow_id: str, setup: Union[RunSetupMetadata, dict]) -> models.Run:
        """Starts a new training run from a specific workflow using the flow ID.

        :param flow_id: The ID of the workflow.
        :param setup: The metadata of the training.
        :return: A msgspec struct containing the
            newly-initialized training run metadata with the following structure:

            .. code-block:: python

                Run(
                    id='run_e2a14cee-eacc-4335-bc95-94c3ee196b04',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    flow_id='flow_64e812a7e47592ef374cbbc2',
                    execution=RunExecution(
                        accelerator=RunAccelerator(name='GPU_L4', count=1),
                        checkpoint=RunCheckpoint(
                            strategy='STRAT_LOWEST_VALIDATION_LOSS',
                            evaluation_interval=220,
                            metric='Loss/total_loss'
                        ),
                        limit=RunLimit(metric='LIM_NONE', value=0)
                    ),
                    status=RunStatus(
                        overview='Creating',
                        message='Training starting.',
                        update_data=1700204316180
                    ),
                    create_date=1701927649302,
                    update_date=1701927649302,
                    features=RunFeatures(
                        preview=True,
                        matrix=True,
                        advanced_evaluation=True
                    ),
                    log_ids=['runlog_UmUyYTE0Y2VlLWVhY2MtNDMzNS1iYzk1LTk0YzNlZTE5NmIwNA']
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.runs.start("flow_63d0f2d5fb1f9189db9b1c4b", {
                    "accelerator": {
                        "name": "GPU_T4",
                        "count": 1
                    },
                    "checkpoint": {
                        "strategy": "STRAT_ALWAYS_SAVE_LATEST",
                        "evaluation_interval": 250
                    },
                    "limit": {
                        "metric": "LIM_NONE",
                        "value": 0
                    },
                    "advanced_evaluation": True,
                })
        """
        assert isinstance(flow_id, str)
        assert isinstance(setup, (RunSetupMetadata, dict))

        if isinstance(setup, dict):
            setup = RunSetupMetadata(**setup)

        return self.requester.POST(
            f"/projects/{self.project_id}/runs",
            request_body={
                "flowId": flow_id,
                "execution": {
                    "accelerator": setup.accelerator.to_json(),
                    "checkpoint": setup.checkpoint.to_json(),
                    "limit": setup.limit.to_json(),
                    "debug": setup.debug,
                },
                "features": {
                    "advancedEvaluation": setup.advanced_evaluation,
                    "preview": setup.preview,
                    "matrix": setup.matrix,
                },
            },
            response_type=models.Run,
        )

    def get_logs(self, log_id: str) -> models.RunLogs:
        """
        Retrieves a specific training log using the log ID.

        :param log_id: The ID of the training log.
        :return: A msgspec struct with the specific
            training log metadata with the following structure:

            .. code-block:: python

                RunLogs(
                    id='runlog_UmUyYTE0Y2VlLWVhY2MtNDMzNS1iYzk1LTk0YzNlZTE5NmIwNA',
                    logs=[
                        {
                            'ev': 'trainingCheckpoint',
                            't': 1700200466147,
                            'pl': {
                                'step': 0,
                                'log': 'Step 0, totalLoss: 6.4055E+01, boxLoss: 5.3928E+00, classificationLoss: 5.2646E+01, distributedFocalLoss: 6.0163E+00.',
                                'totalLoss': 64.055,
                                'boxLoss': 5.3928,
                                'classificationLoss': 52.646,
                                'distributedFocalLoss': 6.0163
                            }
                        }
                    ]
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.runs.get_logs("runlog_63eb212ff0f856bf95085095")
        """
        assert isinstance(log_id, str)
        return self.requester.GET(
            f"/projects/{self.project_id}/runs/logs/{log_id}",
            response_type=models.RunLogs,
        )

    def get_confusion_matrix(self, run_id: str) -> models.RunConfusionMatrix:
        """Retrieves a training confusion matrix using the run ID.

        :param run_id: The ID of the training run.
        :return: A msgspec struct containing the specific
            training matrix json string with the following structure:

            .. code-block:: python

                RunConfusionMatrix(
                    "object" = "confusionMatrix",
                    "data" = "{\"0\":[{\"id\":\"RBC\",\"data\":[{\"x\":\"RBC\",\"y\":0},{\"x\":\"WBC\",\"y\":0},{\"x\":\"Platelets\",\"y\":0},{\"x\":\"boat\",\"y\":0},{\"x\":\"Background\",\"y\":0}]},{\"id\":\"WBC\",\"data\":[{\"x\":\"RBC\",\"y\":0},{\"x\":\"WBC\",\"y\":0},{\"x\":\"Platelets\",\"y\":0},{\"x\":\"boat\",\"y\":0},{\"x\":\"Background\",\"y\":0}]},{\"id\":\"Platelets\",\"data\":[{\"x\":\"RBC\",\"y\":0},{\"x\":\"WBC\",\"y\":0},{\"x\":\"Platelets\",\"y\":0},{\"x\":\"boat\",\"y\":0},{\"x\":\"Background\",\"y\":0}]},{\"id\":\"boat\",\"data\":[{\"x\":\"RBC\",\"y\":0},{\"x\":\"WBC\",\"y\":0},{\"x\":\"Platelets\",\"y\":0},{\"x\":\"boat\",\"y\":0},{\"x\":\"Background\",\"y\":0}]},{\"id\":\"Background\",\"data\":[{\"x\":\"RBC\",\"y\":302},{\"x\":\"WBC\",\"y\":27},{\"x\":\"Platelets\",\"y\":22},{\"x\":\"boat\",\"y\":2},{\"x\":\"Background\",\"y\":0}]}]}"
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")
                project.runs.get_confusion_matrix("run_63eb212ff0f856bf95085095")
        """
        assert isinstance(run_id, str)
        return self.requester.GET(
            f"/projects/{self.project_id}/runs/{run_id}/confusionMatrix",
            response_type=models.RunConfusionMatrix,
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   messages.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   collect all string messages variables
"""

REQUEST_SERVER_MESSAGE = "Communicating with server."
SERVER_COMPLETED_MESSAGE = "Server processing completed."
INVALID_PROJECT_SECRET_MESSAGE = (
    "\nInvalid project secret key, "
    "generate one for your project on Nexus: Advanced/API Management"
)
AUTHENTICATION_MESSAGE = "Authentication succeeded."
AUTHENTICATION_REMINDER_MESSAGE = (
    "\nSecret Key needed, "
    "generate one for your project on Nexus: Advanced/API Management"
)
NO_PROJECT_MESSAGE = (
    "\nMissing authentication, please authenticate with 'datature projects auth'."
)
ASSETS_FOLDER_MESSAGE = "Enter the assets folder path to be uploaded"
ASSETS_GROUPS_MESSAGE = "Enter the asset group name(s), split by ','"
ANNOTATION_FOLDER_MESSAGE = (
    "Enter the folder path containing the annotation files to be uploaded"
)
ANNOTATION_FORMAT_MESSAGE = "Select the annotation file format"
NO_ANNOTATIONS_MESSAGE = (
    "\nNo annotations found in the project. Please annotate assets on Nexus, "
    "or upload annotations with 'datature annotations upload'."
)
NO_ARTIFACTS_MESSAGE = (
    "\nNo artifacts can be downloaded, please start a training first."
)
ARTIFACT_DOWNLOAD_MESSAGE = "Which artifact do you want to download?"
ARTIFACT_MODEL_FORMAT_DOWNLOAD_MESSAGE = "Which model format do you want to download?"
EXPORT_ARTIFACT_WAITING_MESSAGE = (
    "Processing artifact for download, it may take 5-10 minutes.\n"
)
ARTIFACT_MODEL_FOLDER_MESSAGE = "Enter the folder path to save model"
EXPORT_ANNOTATION_FOLDER_MESSAGE = "Enter the folder path to save annotation files"
CHOOSE_GROUP_MESSAGE = "Which asset group do you want to list?"
INVALID_PROJECT_MESSAGE = "\nInvalid project name."
PATH_NOT_EXISTS_MESSAGE = "\nPath does not exist."
NO_ASSETS_GROUP_MESSAGE = "\nNo asset groups exist in this project."
DOWNLOAD_ANNOTATIONS_NORMALIZED_MESSAGE = "Should the annotations be normalized?"
DOWNLOAD_ANNOTATIONS_SPLIT_RATIO_MESSAGE = (
    "Enter the split ratio for this download. [0-100%]"
)
INVALID_SPLIT_RATIO_MESSAGE = "\nInvalid split ratio."
AUTHENTICATION_FAILED_MESSAGE = (
    "\nAuthentication failed, please use 'datature projects auth' again."
)
UNKNOWN_ERROR_SUPPORT_MESSAGE = (
    "\nCommunication failed, contact support at support@datature.io."
)
CONNECTION_ERROR_MESSAGE = "\nConnection failed, please check your network."
UNKNOWN_ERROR_MESSAGE = (
    "\nUnknown error occurred, contact support at support@datature.io."
)
ANNOTATION_DOWNLOAD_MESSAGE = "Processing annotations for download."
ANNOTATION_DOWNLOADED_MESSAGE = "Downloaded annotations."
ACTIVE_PROJECT_MESSAGE = "Your active project is now"
NO_ASSETS_FOUND_MESSAGE = (
    "No allowable assets found in folders, please change the folder path."
)
ASSETS_NIFTI_DIRECTION_CHOICE_MESSAGE = (
    "Select the axis of orientation, "
    "if not provided, we will save videos for each axis. ['x','y','z']"
)
RUNNER_EXISTS_MESSAGE = (
    "A Runner already exists. Multiple runners are currently not supported.\n"
    "If you would like to install a new Runner, "
    "please first uninstall the existing Runner by running `datature runner uninstall`.\n"
    "To view all available Runners, run `datature runner list`."
)
RUNNER_INSTALL_SIZE_MESSAGE = (
    "The Runner will take up approximately 754MB of disk space.\n"
    "To run trainings on your Runner, it is also recommended to have "
    "an additional minimum of 8GB of disk space on your system. Continue?"
)
MAX_RETRY_MESSAGE = "Maximum number of retries reached. Please ensure your details are correct and try again."
RUNNER_NAME_MESSAGE = "Please enter the name for your custom runner:"
EMPTY_RUNNER_NAME_MESSAGE = (
    "Runner name cannot be left blank. Please try again.\n"
    "Ensure to provide a meaningful name for easy identification."
)
DUPLICATE_RUNNER_NAME_MESSAGE = (
    "Runner name already exists. Please enter a unique name for your Runner."
)
SECRET_KEY_MESSAGE = "Please enter your workspace secret key:"
NEW_SECRET_KEY_MESSAGE = "Please enter your new workspace secret key:"
EMPTY_SECRET_KEY_MESSAGE = (
    "Workspace secret key cannot be left blank.\n"
    "Generate one for your workspace on Nexus, under Settings/Key Manager."
)
INVALID_SECRET_KEY_LENGTH_MESSAGE = (
    "Invalid secret key length. Please try again.\n"
    "Ensure to provide the correct secret key for your workspace."
)
ALL_RUNNERS_VALID_SECRET_KEYS_MESSAGE = "All Runners have valid secret keys."
WORKSPACE_ID_MESSAGE = "Please enter your workspace ID:"
EMPTY_WORKSPACE_ID_MESSAGE = (
    "Workspace ID cannot be left blank. Please try again.\n"
    "Double-check your workspace ID and enter it correctly."
)
INVALID_WORKSPACE_ID_LENGTH_MESSAGE = (
    "Invalid workspace ID length. Please try again.\n"
    "Ensure to provide the correct workspace ID for your workspace."
)
NO_RUNNERS_MESSAGE = (
    "No existing Runners found. Please run `datature runner install` to set up a Runner.\n"
    "Ensure your environment meets the necessary requirements before installation."
)
NO_ACTIONABLE_RUNNERS_MESSAGE = "No actionable Runners found to {}. Please run `datature runner list` to view Runner statuses."
UNINSTALL_RUNNERS_CONFIRMATION_MESSAGE = (
    "Are you sure you want to uninstall selected Runner(s)? This action is irreversible.\n"
    "Uninstalling will remove all associated data and configurations."
)
KILL_RUNS_CONFIRMATION_MESSAGE = (
    "There may be ongoing runs associated with Runner '{}' [{}] that will be killed.\n"
    "Ensure to save any necessary data before proceeding.\n"
    "Do you wish to continue?"
)
VALID_SECRET_KEY_MESSAGE = (
    "Secret key for Runner '{}' [{}] is valid, reauthentication is not necessary."
)
INVALID_SECRET_KEYS_MESSAGE = (
    "One or more workspace secret keys are invalid. "
    "A user in your workspace may have generated new keys.\n"
    "Please run `datature runner reauth` to reauthenticate, "
    "and double-check your secret keys for accuracy."
)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   commands.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   CLI Commands & Arguments
"""

import sys
from argparse import ArgumentParser, _SubParsersAction
from datetime import datetime
from typing import Optional

from datature import nexus

# pylint: disable=W0212


class DatatureArgumentParser(ArgumentParser):
    """Custom ArgumentParser to handle custom error messages."""

    def error(self, message):
        if "unrecognized arguments: help" in message:
            sys.stderr.write(
                "Error: unrecognized argument 'help'. Did you mean '--help'?\n"
            )
        else:
            sys.stderr.write(f"Error: {message}\n")
            sys.stderr.write("Use '--help' for more information.\n")
        sys.exit(2)


class Commands:
    """All Datature CLI commands."""

    def __init__(self) -> None:
        self.parser = DatatureArgumentParser(
            prog="datature",
            description="Command line tool for interaction with Datature Nexus.",
        )

        self.parser.add_argument(
            "-v", "--version", action="version", version=f"%(prog)s {nexus.__version__}"
        )

        self.subparsers = self.parser.add_subparsers(dest="command")
        self._add_project_parser()
        self._add_asset_parser()
        self._add_annotation_parser()
        self._add_artifact_parser()
        self._add_runner_parser()
        self._add_batch_parser()
        self._add_help_parser()

    def _add_project_parser(self):
        """Project level parser."""
        projects = self.subparsers.add_parser(
            "projects",
            help="Projects management.",
            description="datature projects - auth/list/select project from saved project.",
        )
        projects_action = projects.add_subparsers(dest="action")

        projects_action.add_parser("auth", help="Authenticate and save the project.")
        projects_action.add_parser(
            "select", help="Select the project from saved projects."
        )
        projects_action.add_parser("list", help="List the saved projects.")

        projects_action.add_parser(
            "help", help="Show this help message and exit.", add_help=False
        )

    def _add_asset_parser(self):
        """Asset level parser."""
        assets = self.subparsers.add_parser(
            "assets",
            help="Assets management.",
            description="datature assets - upload/group assets.",
        )
        assets_action = assets.add_subparsers(dest="action")

        assets_action.add_parser(
            "upload",
            help="Bulk update assets.",
        )

        assets_action.add_parser(
            "groups",
            help="List assets group details.",
        )

        assets_action.add_parser(
            "help", help="Show this help message and exit.", add_help=False
        )

    def _add_annotation_parser(self):
        """Annotation level parser."""
        annotations = self.subparsers.add_parser(
            "annotations",
            help="Annotations management.",
            description="datature annotations - upload/download annotations.",
        )
        annotations_action = annotations.add_subparsers(dest="action")

        annotations_action.add_parser(
            "upload",
            help="Bulk upload annotations from file.",
        )

        annotations_action.add_parser(
            "download",
            help="Bulk download annotations to file.",
        )

        annotations_action.add_parser(
            "help", help="Show this help message and exit.", add_help=False
        )

    def _add_artifact_parser(self):
        """Artifact level parser."""
        artifacts = self.subparsers.add_parser(
            "artifacts",
            help="Artifacts management.",
            description="datature artifacts - download artifact models.",
        )
        artifacts_action = artifacts.add_subparsers(dest="action")

        artifacts_action.add_parser(
            "download",
            help="Download artifact model.",
        )

        artifacts_action.add_parser(
            "help", help="Show this help message and exit.", add_help=False
        )

    def _add_runner_parser(self):
        """Runner level parser."""
        runner = self.subparsers.add_parser(
            "runner",
            help="Runner management.",
            description="datature runner - setup and manage custom runners.",
        )
        details_parser = ArgumentParser(add_help=False)
        details_parser.add_argument(
            "name_or_hash",
            type=str,
            nargs="?",
            help="Runner name or hash.",
        )
        runner_action = runner.add_subparsers(dest="action")
        runner_action.add_parser("install", help="Install the custom runner.")
        runner_action.add_parser(
            "uninstall", parents=[details_parser], help="Uninstall the custom runner."
        )
        runner_action.add_parser(
            "suspend", parents=[details_parser], help="Suspend the custom runner."
        )
        runner_action.add_parser(
            "resume", parents=[details_parser], help="Resume the custom runner."
        )
        runner_action.add_parser(
            "restart", parents=[details_parser], help="Restart the custom runner."
        )
        runner_action.add_parser(
            "reauth", parents=[details_parser], help="Reauthenticate the custom runner."
        )
        runner_logs = runner_action.add_parser(
            "logs",
            parents=[details_parser],
            help=(
                "Dump the log of the custom runner. "
                "Use -r to specify the run hash. Use -o to specify the output log file path."
            ),
        )
        runner_logs.add_argument(
            "-r",
            "--run",
            action="store",
            help="Run Hash",
            default=None,
        )
        runner_logs.add_argument(
            "-o",
            "--output",
            action="store",
            help="Output log file path",
            default=None,
        )
        runner_action.add_parser(
            "status",
            parents=[details_parser],
            help="Display the status of the custom runner.",
        )
        runner_action.add_parser("list", help="List all the custom runners.")
        runner_action.add_parser(
            "help", help="Show this help message and exit.", add_help=False
        )

    def _add_batch_parser(self):
        """Batch level parser."""
        batch = self.subparsers.add_parser(
            "batch",
            help="Batch job management.",
            description="datature batch jobs - manage batch jobs.",
        )
        entity = batch.add_subparsers(dest="entity")
        entity.add_parser(
            "help", help="Show this help message and exit.", add_help=False
        )
        details_parser = ArgumentParser(add_help=False)
        details_parser.add_argument(
            "id",
            type=str,
            nargs="?",
            help="ID of the entry to get details for.",
        )

        def _add_jobs_parser():
            jobs = entity.add_parser("jobs", help="Manage batch jobs.")
            jobs_action = jobs.add_subparsers(dest="action")
            jobs_action.add_parser(
                "create",
                help="Create a new batch job.",
            )
            jobs_action.add_parser(
                "get",
                parents=[details_parser],
                help="Get details of a batch job.",
            )
            jobs_action.add_parser(
                "list",
                help="List batch jobs.",
            )
            jobs_action.add_parser(
                "wait-until-done",
                parents=[details_parser],
                help="Wait for a batch job to complete.",
            )
            jobs_action.add_parser(
                "cancel",
                parents=[details_parser],
                help="Cancel a batch job.",
            )
            jobs_action.add_parser(
                "delete",
                parents=[details_parser],
                help="Delete a batch job.",
            )

            job_logs = jobs_action.add_parser(
                "logs",
                parents=[details_parser],
                help="Get logs of a batch job.",
            )
            job_logs.add_argument(
                "--max_entries",
                type=int,
                default=0,
                help="Maximum number of log entries to return.",
            )
            job_logs.add_argument(
                "--since",
                type=int,
                default=0,
                help="Only return logs after this timestamp.",
            )
            job_logs.add_argument(
                "--until",
                type=int,
                default=int(datetime.now().timestamp()) * 1000,
                help="Only return logs before this timestamp.",
            )
            job_logs.add_argument(
                "--level",
                type=str,
                default="Info",
                help="Only return logs at this level or higher.",
            )
            job_logs.add_argument(
                "--output",
                type=str,
                default="",
                help="Output log to file.",
            )

        def _add_webhooksecrets_parser():
            webhooksecrets = entity.add_parser(
                "webhooksecrets", help="Manage webhook secrets."
            )
            webhooksecrets_action = webhooksecrets.add_subparsers(dest="action")
            webhooksecrets_action.add_parser(
                "create",
                help="Create a new webhook secret.",
            )

        def _add_webhooks_parser():
            webhooks = entity.add_parser("webhooks", help="Manage webhooks.")
            webhooks_action = webhooks.add_subparsers(dest="action")
            webhooks_action.add_parser(
                "create",
                help="Create a new webhook.",
            )
            webhooks_action.add_parser(
                "get",
                parents=[details_parser],
                help="Get details of a webhook.",
            )
            webhooks_action.add_parser(
                "list",
                help="List webhooks.",
            )
            webhooks_action.add_parser(
                "update",
                parents=[details_parser],
                help="Update a webhook.",
            )
            webhooks_action.add_parser(
                "update-secret",
                parents=[details_parser],
                help="Update the secret of a webhook.",
            )
            webhooks_action.add_parser(
                "delete",
                parents=[details_parser],
                help="Delete a webhook.",
            )
            webhooks_action.add_parser(
                "test",
                parents=[details_parser],
                help="Test a webhook by sending a test payload.",
            )

        def _add_datasets_parser():
            datasets = entity.add_parser("datasets", help="Manage datasets.")
            datasets_action = datasets.add_subparsers(dest="action")
            datasets_action.add_parser(
                "create",
                help="Create a new dataset.",
            )
            datasets_action.add_parser(
                "get",
                parents=[details_parser],
                help="Get details of a dataset.",
            )
            datasets_action.add_parser(
                "list",
                help="List datasets.",
            )
            datasets_action.add_parser(
                "delete",
                parents=[details_parser],
                help="Delete a dataset.",
            )

        _add_jobs_parser()
        _add_webhooksecrets_parser()
        _add_webhooks_parser()
        _add_datasets_parser()

    def _add_help_parser(self):
        """Help level parser."""
        help_parser = self.subparsers.add_parser(
            "help",
            description="Show this help message and exit.",
        )
        help_action = help_parser.add_subparsers(dest="action")

        command_names = ["project", "asset", "annotation", "artifact", "batch"]
        for command_name in command_names:
            help_action.add_parser(command_name)

    def parse_args(self) -> ArgumentParser:
        """
        Parses and validates the CLI commands.

        :return: The parser to use.
        """
        args = self.parser.parse_args()

        if not args.command:
            self.print_help()
            sys.exit()

        return args

    def print_help(self, subparser: Optional[str] = None):
        """
        Prints the help information.

        :param subparser: The name of subparser.
        :return: None
        """
        parser = self.parser

        if subparser:
            parser = next(
                action.choices[subparser]
                for action in parser._actions
                if isinstance(action, _SubParsersAction) and subparser in action.choices
            )

        parser.print_help()



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_asset_upload_session.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Asset Upload Session API Test Cases
"""
# pylint: disable=W0703,R0913,W0613,W0212

import json
import unittest
from unittest.mock import MagicMock, patch

import msgspec

from datature.nexus import models
from datature.nexus.api.asset.upload_session import UploadSession
from datature.nexus.client_context import ClientContext
from datature.nexus.error import Error
from tests.unit.fixture.data import projects_fixture, upload_session_fixture


class TestAssetUploadSession(unittest.TestCase):
    """Datature Asset Upload Session API Resource Test Cases."""

    def __init__(self, *args, **kwargs):
        """init global variables."""
        unittest.TestCase.__init__(self, *args, **kwargs)

        context = ClientContext("secret_key", endpoint="http://localhost:8080")
        context.project_id = "project_id"
        context.requester._request = MagicMock()
        self.context = context

    def test_add_with_file_not_exist(self):
        """Test add asset to upload with ."""
        upload_session = UploadSession(self.context)

        try:
            upload_session.add_path("assetPath")
        except Exception as exception:
            assert isinstance(exception, Error)

    @patch("datature.nexus.api.asset.upload_session.config")
    @patch("datature.nexus.api.asset.upload_session.filetype")
    @patch("datature.nexus.api.asset.upload_session.struct")
    @patch("datature.nexus.api.asset.upload_session.crc32c")
    @patch("datature.nexus.api.asset.upload_session.path")
    @patch("datature.nexus.api.asset.upload_session.open")
    def test_add_with_duplicated_file(
        self, patch_open, patch_path, crc32c, struct, filetype, patch_config
    ):
        """Test add asset to upload with duplicated file."""
        patch_config.ASSET_UPLOAD_SESSION_BATCH_SIZE = 100

        upload_session = UploadSession(self.context)
        upload_session.file_name_map = {"assetName": {"path": "file_path"}}

        struct.unpack.return_value = [-384617082]
        patch_path.basename.return_value = "assetName"
        patch_path.getsize.return_value = 5613

        mock_guess = MagicMock()
        mock_guess.mime = "image/jpeg"
        filetype.guess.return_value = mock_guess

        try:
            with upload_session as upload_session:
                upload_session.add_path("assetPath")
        except Exception as exception:
            assert isinstance(exception, Error)

    @patch("datature.nexus.api.asset.upload_session.config")
    @patch("datature.nexus.api.asset.upload_session.filetype")
    @patch("datature.nexus.api.asset.upload_session.path")
    @patch("datature.nexus.api.asset.upload_session.struct")
    @patch("datature.nexus.api.asset.upload_session.crc32c")
    @patch("datature.nexus.api.asset.upload_session.open")
    def test_add_with_file_not_supported(
        self, patch_open, crc32c, struct, path, filetype, patch_config
    ):
        """Test add asset to upload with file not supported."""
        patch_config.ASSET_UPLOAD_SESSION_BATCH_SIZE = 100

        upload_session = UploadSession(self.context)

        struct.unpack.return_value = [-384617082]
        path.basename.return_value = "assetName"
        path.getsize.return_value = 5613

        mock_guess = MagicMock()
        mock_guess.mime = ""
        filetype.guess.return_value = mock_guess

        try:
            with upload_session as upload_session:
                upload_session.add_path("assetPath")
        except Exception as exception:
            assert isinstance(exception, Error)

    @patch("datature.nexus.api.asset.upload_session.config")
    @patch("datature.nexus.api.asset.upload_session.filetype")
    @patch("datature.nexus.api.asset.upload_session.path")
    @patch("datature.nexus.api.asset.upload_session.struct")
    @patch("datature.nexus.api.asset.upload_session.crc32c")
    @patch("datature.nexus.api.asset.upload_session.open")
    def test_add_with_file(
        self, patch_open, crc32c, struct, path, filetype, patch_config
    ):
        """Test add asset to upload with file."""
        patch_config.ASSET_UPLOAD_SESSION_BATCH_SIZE = 100
        upload_session = UploadSession(self.context)

        struct.unpack.return_value = [-384617082]
        path.basename.return_value = "assetName"
        path.getsize.return_value = 5613

        mock_guess = MagicMock()
        mock_guess.mime = "image/jpeg"
        filetype.guess.return_value = mock_guess

        upload_session.add_path("assetPath")

    def test_start_with_empty_assets(self):
        """Test upload with empty assets ."""
        upload_session = UploadSession(self.context, ["main"])
        try:
            with upload_session as upload_session:
                upload_session.add_path("assetPath")
        except Exception as exception:
            assert isinstance(exception, Error)

    @patch("datature.nexus.api.asset.upload_session.config")
    @patch("datature.nexus.api.asset.upload_session.open")
    def test_start(self, patch_open, patch_config):
        """Test start upload with empty assets ."""
        patch_config.ASSET_UPLOAD_SESSION_BATCH_SIZE = 100

        upload_session = UploadSession(self.context, ["main"], background=True)
        upload_session._upload_assets = MagicMock()

        upload_session.assets = [
            {
                "filename": "test.jpeg",
                "mime": "image/jpeg",
                "size": 5613,
                "crc32c": -384617082,
            }
        ]
        upload_session.file_name_map = {"test.jpeg": {"path": "file_path"}}

        upload_session._context.requester._request.return_value = msgspec.json.decode(
            json.dumps(upload_session_fixture.upload_assets_response),
            type=models.UploadSession,
        )

        upload_session.assets = [
            {
                "filename": "test.jpeg",
                "mime": "image/jpeg",
                "size": 5613,
                "crc32c": -384617082,
            }
        ]

        with upload_session as upload_session:
            pass

    @patch("datature.nexus.api.asset.upload_session.open")
    def test_start_with_background(self, patch_open):
        """Test start upload with wait server process."""
        upload_session = UploadSession(self.context, ["main"])

        upload_session.operation.wait_until_done = MagicMock()
        upload_session._upload_assets = MagicMock()

        upload_session.assets = [
            {
                "filename": "test.jpeg",
                "mime": "image/jpeg",
                "size": 5613,
                "crc32c": -384617082,
            }
        ]
        upload_session.file_name_map = {"test.jpeg": {"path": "file_path"}}

        upload_session._context.requester._request.return_value = msgspec.json.decode(
            json.dumps(upload_session_fixture.upload_assets_response),
            type=models.UploadSession,
        )

        upload_session.assets = [
            {
                "filename": "test.jpeg",
                "mime": "image/jpeg",
                "size": 5613,
                "crc32c": -384617082,
            }
        ]

        with upload_session as upload_session:
            pass

        upload_session.operation.wait_until_done.assert_called()

    def test_get_operation_ids(self):
        """Test get_operation_ids function."""
        upload_session = UploadSession(self.context, ["main"])
        upload_session.operation_ids = ["id1", "id2"]

        operation_ids = upload_session.get_operation_ids()

        assert operation_ids == ["id1", "id2"]

    def test_import_session_length(self):
        """Test len(upload_session) function."""
        upload_session = UploadSession(self.context, ["main"])
        upload_session.file_name_map = {
            "filename": {
                "path": "path",
            }
        }

        assert len(upload_session) == 1



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   client.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   SDK Client module
"""
# pylint: disable=C0103

import copy
import os
from abc import ABC
from typing import List

from datature.nexus import models
from datature.nexus.api.project import Project
from datature.nexus.client_context import ClientContext


class Client(ABC):
    """Initialize the Datature client with a secret key and an API endpoint.

    Parameters:
        secret_key (str): The secret key used for authentication with the Datature API.
        endpoint (str): The base URL for the Datature API.
            Defaults to the value of the DATATURE_API_BASE_URL environment variable,
            or "https://api.datature.io" if the environment variable is not set.
    """

    _context: ClientContext

    def __init__(
        self,
        secret_key: str,
        endpoint: str = os.getenv("DATATURE_API_BASE_URL", "https://api.datature.io"),
    ):
        """Initialize"""
        self._context = ClientContext(secret_key, endpoint)

    def get_info(self) -> models.Workspace:
        """Retrieve this workspace.

        :return: A msgspec struct containing the workspace information with the following structure:

            .. code-block:: python

                Workspace(
                    id='ws_1c8aab980f174b0296c7e35e88665b13',
                    name="Raighne's Workspace",
                    owner='user_6323fea23e292439f31c58cd',
                    tier='Developer',
                    create_date=1701927649302
                )

        :example:

            .. code-block:: python

                from datature.nexus import Client

                client = Client("5aa41e8ba........")
                client.get_info()
        """
        return self._context.requester.GET("/workspace", response_type=models.Workspace)

    def list_projects(self) -> List[Project]:
        """List all projects under this workspace.

        :return: A List of project class instances.
        :example:

            .. code-block:: python

                from datature.nexus import Client

                projects = Client("5aa41e8ba........").list_projects()

                for project in projects:
                    print(project.get_info())
        """
        projects = self._context.requester.GET(
            "/workspace/projects", response_type=models.Projects
        )

        projects_res = []
        for project in projects:
            project_context = copy.deepcopy(self._context)
            project_context.project_id = project.id
            projects_res.append(Project(project_context))

        return projects_res

    def get_project(self, project_id: str) -> Project:
        """Return a project object with a separate context.

        :param project_id: The project ID.
        :return: A class instance of the project.
        :example:

            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")
        """
        assert isinstance(project_id, str)

        if not project_id.startswith("proj_"):
            project_id = f"proj_{project_id}"

        project_context = copy.deepcopy(self._context)
        project_context.project_id = project_id

        return Project(project_context)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   webhook.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Webhook API
"""

from typing import Union

from datature.nexus import models
from datature.nexus.api.batch import models as batch_models
from datature.nexus.api.batch.types import WebhookMetadata, WebhookSpec
from datature.nexus.api.types import Pagination
from datature.nexus.client_context import RestContext


class Webhook(RestContext):
    """Datature Webhook API Resource."""

    def create(
        self, name: str, webhook_spec: Union[WebhookSpec, dict, None] = None
    ) -> batch_models.WebhookModel:
        """Create a new webhook.

        :param name: The name of the webhook.
        :param webhook_spec: The webhook specification.
        :return: A msgspec struct containing the webhook metadata with the following structure:

            .. code-block:: python

                WebhookModel(
                    id='webhook_f7d8aec2-7e2b-4d2c-a103-c8dd575c29c7',
                    object='webhook',
                    project_id='proj_ca5fe71e7592bbcf7705ea36e4f29ed4',
                    name='aws-lambda',
                    endpoint='https://lambda-url.ap-southeast-1.on.aws/',
                    retries=WebhookRetries(
                        max_retries=3,
                        max_retry_delay=15000
                    ),
                    create_date=1723633671620,
                    update_date=1723736909980
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                from datature.nexus.utils import utils
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                webhook_secret = utils.generate_webhook_secret()

                project.batch.webhooks.create(
                    name="aws-lambda",
                    webhook_spec={
                        "endpoint": "https://lambda-url.ap-southeast-1.on.aws/",
                        "retries": {
                            "max_retries": 3,
                            "max_retry_delay": 15000
                        },
                        "secret": {
                            "contents": webhook_secret,
                        },
                    }
                )
        """
        assert isinstance(webhook_spec, (WebhookSpec, dict, type(None)))
        if isinstance(WebhookSpec, dict):
            webhook_spec = WebhookSpec(**webhook_spec)
        webhook_options = WebhookMetadata(name=name, spec=webhook_spec)

        return self.requester.POST(
            f"/projects/{self.project_id}/batch/webhooks",
            request_body=webhook_options.to_json(),
            response_type=batch_models.WebhookModel,
        )

    def delete(self, webhook_id: str) -> models.DeleteResponse:
        """Delete a webhook.

        :param webhook_id: The webhook ID.
        :return: A msgspec struct containing the delete response with the following structure:

            .. code-block:: python

                DeleteResponse(
                    id='webhook_f7d8aec2-7e2b-4d2c-a103-c8dd575c29c7',
                    deleted=True
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.batch.webhooks.delete("webhook_f7d8aec2-7e2b-4d2c-a103-c8dd575c29c7")
        """
        assert isinstance(webhook_id, str)
        return self.requester.DELETE(
            f"/projects/{self.project_id}/batch/webhooks/{webhook_id}",
            response_type=models.DeleteResponse,
        )

    def get(self, webhook_id: str) -> batch_models.WebhookModel:
        """Get a webhook.

        :param webhook_id: The webhook ID.
        :return: A msgspec struct containing the webhook metadata with the following structure:

            .. code-block:: python

                WebhookModel(
                    id='webhook_f7d8aec2-7e2b-4d2c-a103-c8dd575c29c7',
                    object='webhook',
                    project_id='proj_ca5fe71e7592bbcf7705ea36e4f29ed4',
                    name='localhost',
                    endpoint='https://lambda-url.ap-southeast-1.on.aws/',
                    retries=WebhookRetries(
                        max_retries=3,
                        max_retry_delay=15000
                    ),
                    create_date=1723633671620,
                    update_date=1723736909980
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.batch.webhooks.get("webhook_f7d8aec2-7e2b-4d2c-a103-c8dd575c29c7")
        """
        assert isinstance(webhook_id, str)
        return self.requester.GET(
            f"/projects/{self.project_id}/batch/webhooks/{webhook_id}",
            response_type=batch_models.WebhookModel,
        )

    def list(
        self,
        pagination: Union[Pagination, dict, None] = None,
    ) -> models.PaginationResponse[batch_models.WebhookModel]:
        """List all webhooks.

        :param pagination: The pagination options.
        :return: A msgspec struct containing the pagination response with the following structure:

            .. code-block:: python

                PaginationResponse(
                    next_page=None,
                    prev_page=None,
                    data=[
                        WebhookModel(
                            id='webhook_f7d8aec2-7e2b-4d2c-a103-c8dd575c29c7',
                            object='webhook',
                            project_id='proj_ca5fe71e7592bbcf7705ea36e4f29ed4',
                            name='localhost',
                            endpoint='https://lambda-url.ap-southeast-1.on.aws/',
                            retries=WebhookRetries(
                                max_retries=3,
                                max_retry_delay=15000
                            ),
                            create_date=1723633671620,
                            update_date=1723736909980
                        )
                    ]
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.batch.webhooks.list()
        """
        assert isinstance(pagination, (Pagination, dict, type(None)))

        if isinstance(pagination, dict):
            pagination = Pagination(**pagination)
        if pagination is None:
            pagination = Pagination()

        return self.requester.GET(
            f"/projects/{self.project_id}/batch/webhooks",
            query={**pagination.to_json()},
            response_type=models.PaginationResponse[batch_models.WebhookModel],
        )

    def update(
        self, webhook_id: str, webhook_spec: Union[WebhookSpec, dict, None]
    ) -> batch_models.WebhookModel:
        """Update a webhook.

        :param webhook_id: The webhook ID.
        :param webhook_spec: The webhook specification.
        :return: A msgspec struct containing the webhook metadata with the following structure:

            .. code-block:: python

                WebhookModel(
                    id='webhook_f7d8aec2-7e2b-4d2c-a103-c8dd575c29c7',
                    object='webhook',
                    project_id='proj_ca5fe71e7592bbcf7705ea36e4f29ed4',
                    name='localhost',
                    endpoint='https://lambda-url.ap-southeast-1.on.aws/',
                    retries=WebhookRetries(
                        max_retries=3,
                        max_retry_delay=15000
                    ),
                    create_date=1723633671620,
                    update_date=1723736909980
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.batch.webhooks.update(
                    webhook_id="webhook_f7d8aec2-7e2b-4d2c-a103-c8dd575c29c7",
                    webhook_spec={
                        "endpoint": "https://lambda-url.ap-southeast-1.on.aws/",
                        "retries": {
                            "max_retries": 3,
                            "max_retry_delay": 15000
                        }
                    }
                )
        """
        assert isinstance(webhook_spec, (WebhookSpec, dict, type(None)))
        if isinstance(webhook_spec, dict):
            webhook_spec = WebhookSpec(**webhook_spec)
            webhook_spec.secret = None
        webhook_options = WebhookMetadata(spec=webhook_spec)

        return self.requester.PATCH(
            f"/projects/{self.project_id}/batch/webhooks/{webhook_id}",
            request_body=webhook_options.to_json(),
            response_type=batch_models.WebhookModel,
        )

    def update_secret(self, webhook_id: str, secret: str) -> batch_models.WebhookModel:
        """Update a webhook secret.

        :param webhook_id: The webhook ID.
        :param secret: The webhook secret.
        :return: A msgspec struct containing the webhook metadata with the following structure:

            .. code-block:: python

                WebhookModel(
                    id='webhook_f7d8aec2-7e2b-4d2c-a103-c8dd575c29c7',
                    object='webhook',
                    project_id='proj_ca5fe71e7592bbcf7705ea36e4f29ed4',
                    name='localhost',
                    endpoint='https://lambda-url.ap-southeast-1.on.aws/',
                    retries=WebhookRetries(
                        max_retries=3,
                        max_retry_delay=15000
                    ),
                    create_date=1723633671620,
                    update_date=1723736909980
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                from datature.nexus.utils import utils
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                webhook_secret = utils.generate_webhook_secret()

                project.batch.webhooks.update_secret(
                    webhook_id="webhook_f7d8aec2-7e2b-4d2c-a103-c8dd575c29c7",
                    secret=webhook_secret,
                )
        """
        assert isinstance(secret, str)
        return self.requester.PUT(
            f"/projects/{self.project_id}/batch/webhooks/{webhook_id}/secret",
            request_body={"contents": secret},
            response_type=batch_models.WebhookModel,
        )

    def test(self, webhook_id: str) -> batch_models.WebhookTestResponse:
        """Test a webhook by sending a sample payload.

        :param webhook_id: The webhook ID.
        :return: A msgspec struct containing the webhook test response with the following structure:

            .. code-block:: python

                WebhookTestResponse(
                    status='Ok',
                    response_code=204,
                    latency_ms=648,
                    attempt_count=1,
                    body=None,
                    reason=None
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.batch.webhooks.test("webhook_f7d8aec2-7e2b-4d2c-a103-c8dd575c29c7")
        """
        assert isinstance(webhook_id, str)
        return self.requester.POST(
            f"/projects/{self.project_id}/batch/webhooks/{webhook_id}:test",
            response_type=batch_models.WebhookTestResponse,
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_deploy.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Deploy API Test Cases
"""

import unittest
from unittest.mock import ANY, MagicMock

from datature.nexus.api.deployment import Deployment
from datature.nexus.client_context import ClientContext

# pylint: disable=W0212


class TestDeploy(unittest.TestCase):
    """Datature Deploy API Resource Test Cases."""

    def __init__(self, *args, **kwargs):
        """init global variables."""
        unittest.TestCase.__init__(self, *args, **kwargs)

        context = ClientContext("secret_key", endpoint="http://localhost:8080")
        context.project_id = "project_id"
        context.requester._request = MagicMock()

        self.deployment = Deployment(context)

    def test_list(self):
        """Test list all deployments."""
        self.deployment.list()

        self.deployment._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/deployments",
            ANY,
            query=None,
            request_params=None,
        )

    def test_get(self):
        """Test get a deployment."""
        self.deployment.get("deploy_id")

        self.deployment._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/deployments/deploy_id",
            ANY,
            query=None,
            request_params=None,
        )

    def test_delete(self):
        """Test delete a deployment."""
        self.deployment.delete("deploy_id")

        self.deployment._context.requester._request.assert_called_once_with(
            "DELETE", "/projects/project_id/deployments/deploy_id", ANY
        )

    def test_create(self):
        """Test create a deployment."""
        self.deployment.create(
            {
                "name": "name",
                "model_id": "model_id",
                "replicas": 1,
                "resources": {"GPU_T4": 1},
                "options": {"evaluation_strategy": "evaluation_strategy"},
            }
        )

        self.deployment._context.requester._request.assert_called_once_with(
            "POST",
            "/projects/project_id/deployments",
            ANY,
            request_body={
                "name": "name",
                "modelId": "model_id",
                "replicas": 1,
                "resources": {"GPU_T4": 1},
                "options": {"evaluationStrategy": "evaluation_strategy"},
            },
            query=None,
            request_params=None,
        )

    def test_update(self):
        """Test update a deployment."""
        self.deployment.update("deploy_id", {"name": "name", "replicas": 2})

        self.deployment._context.requester._request.assert_called_once_with(
            "PATCH",
            "/projects/project_id/deployments/deploy_id",
            ANY,
            request_body={"name": "name", "replicas": 2},
            query=None,
            request_params=None,
        )

    def test_create_version(self):
        """Test create a deployment version."""
        self.deployment.create_version("deploy_id", "name", "artifact_id")

        self.deployment._context.requester._request.assert_called_once_with(
            "POST",
            "/projects/project_id/deployments/deploy_id/versions",
            ANY,
            request_body={
                "versionTag": "name",
                "artifactId": "artifact_id",
            },
            query=None,
            request_params=None,
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   client_context.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   SDK Client module
"""
# pylint: disable=R0903

from datature.nexus.requester import Requester


class ClientContext:
    """
    Client context for Datature SDK.
    """

    secret_key: str
    requester: Requester
    _project_id: str = None

    def __init__(self, secret_key, endpoint: str):
        self.secret_key = secret_key
        self.requester = Requester(secret_key, endpoint)

    @property
    def project_id(self):
        """The project id.

        :param value: The project id.
        :type: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, value):
        """The project id.

        :param value: The project id.
        :type: str
        """
        self._project_id = value


class RestContext:
    """Datature REST Context."""

    _context: ClientContext

    def __init__(self, context: ClientContext):
        """Create a REST Context.

        :param context: REST context used to make REST calls.
        """
        self.requester = context.requester
        self.project_id = context.project_id
        self._context = context



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   custom_runner.py
@Author  :   Wei Loon Cheng, Kai Xuan Lee
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Custom runner module.
"""

# pylint: disable=W0718,E1120

import json
import queue
import shlex
import subprocess
import sys
import time
from datetime import datetime
from pathlib import Path
from typing import List

import requests

from datature.nexus.cli.runner import commands as COMMANDS
from datature.nexus.cli.runner import common
from datature.nexus.cli.runner import consts as CONSTS
from datature.nexus.cli.runner import messages as MESSAGES
from datature.nexus.cli.runner import utils
from datature.nexus.cli.runner.config import RunnerConfig, RunStatus
from datature.nexus.config import REQUEST_TIME_OUT_SECONDS
from datature.nexus.error import Error

try:
    import kubernetes as k8s
except ImportError as exc:
    if not common.install_pip_package("kubernetes"):
        raise Error(
            MESSAGES.INSTALL_PIP_PACKAGE_ERROR_MESSAGE.format(package_name="kubernetes")
        ) from exc
    import kubernetes as k8s


class CustomRunner:
    """Custom Runner class."""

    def __init__(self, **kwargs):
        CONSTS.RUNNER_INIT_ROOT_DIR.mkdir(parents=True, exist_ok=True)
        self._runner_config = RunnerConfig.read_from_file(
            kwargs.get("runner_name_or_hash", None)
        )

        if not self._runner_config.id:
            self._runner_config.id = utils.create_runner_id(self._runner_config)
            self._runner_config.write_to_file()

        k8s_config_data = subprocess.check_output(
            shlex.split(COMMANDS.MICROK8S_CONFIG_COMMAND), stderr=subprocess.STDOUT
        )
        CONSTS.K8S_CONFIG_DIR.mkdir(parents=True, exist_ok=True)
        CONSTS.K8S_CONFIG_FILE_PATH.write_bytes(k8s_config_data)

        k8s_config = k8s.client.Configuration()
        k8s.config.load_kube_config(
            config_file=str(CONSTS.K8S_CONFIG_FILE_PATH),
            client_configuration=k8s_config,
        )
        with k8s.client.ApiClient(k8s_config) as api_client:
            self._k8s_core_api_instance = k8s.client.CoreV1Api(api_client)
            self._k8s_apps_api_instance = k8s.client.AppsV1Api(api_client)
            self._k8s_rbac_api_instance = k8s.client.RbacAuthorizationV1Api(api_client)
            self._k8s_batch_api_instance = k8s.client.BatchV1Api(api_client)
            self._k8s_custom_api_instance = k8s.client.CustomObjectsApi(api_client)

    def install(self):
        """Install custom runner."""
        try:
            utils.register_runner(self._runner_config)
            utils.create_runner_namespace(
                self._runner_config, self._k8s_core_api_instance
            )
            utils.get_runner_metadata(self._runner_config)
            utils.apply_config_map(self._runner_config, self._k8s_core_api_instance)
            utils.apply_service_account(
                self._runner_config,
                self._k8s_core_api_instance,
                self._k8s_rbac_api_instance,
            )
            utils.setup_docker_registry_credentials(
                self._runner_config, self._k8s_core_api_instance
            )
            utils.apply_deployment(self._runner_config, self._k8s_apps_api_instance)
            utils.initialize_runner(
                self._runner_config,
                self._k8s_apps_api_instance,
                self._k8s_custom_api_instance,
            )
            self._cleanup()
        except KeyboardInterrupt:
            print(MESSAGES.KEYBOARD_INTERRUPT_MESSAGE)
            print(MESSAGES.INSTALL_CLEANUP_MESSAGE)
            self._soft_uninstall()
            sys.exit(1)
        except Error:
            common.log_errors()
            self._soft_uninstall()
            sys.exit(1)
        except Exception:
            common.log_errors(unknown=True)
            self._soft_uninstall()
            sys.exit(1)

        print(
            MESSAGES.INSTALL_RUNNER_SUCCESS_MESSAGE.format(
                self._runner_config.workspace_id
            )
        )

    def _soft_uninstall(self):
        """Soft uninstall custom runner. Cleans up broken installations."""
        try:
            requests.delete(
                f"{CONSTS.NEXUS_ENDPOINT}/workspaces/ws_{self._runner_config.workspace_id}"
                f"/runners/{self._runner_config.id}",
                headers={
                    "Secret-Key": self._runner_config.secret_key,
                    "Content-Type": "application/json",
                },
                timeout=REQUEST_TIME_OUT_SECONDS,
            )
            self._k8s_core_api_instance.delete_collection_namespaced_secret(
                self._runner_config.kubectl_namespace
            )
            self._k8s_apps_api_instance.delete_collection_namespaced_deployment(
                self._runner_config.kubectl_namespace
            )
            self._k8s_rbac_api_instance.delete_collection_namespaced_role_binding(
                self._runner_config.kubectl_namespace
            )
            self._k8s_rbac_api_instance.delete_collection_namespaced_role(
                self._runner_config.kubectl_namespace
            )
            self._k8s_core_api_instance.delete_collection_namespaced_service_account(
                self._runner_config.kubectl_namespace
            )
            self._k8s_core_api_instance.delete_collection_namespaced_config_map(
                self._runner_config.kubectl_namespace
            )
            self._k8s_core_api_instance.delete_namespace(
                self._runner_config.kubectl_namespace
            )
        except k8s.client.rest.ApiException as exc:
            if exc.status != 404:
                raise Error(
                    MESSAGES.UNINSTALL_RUNNER_ERROR_MESSAGE.format(
                        exc.body, self._runner_config.name, self._runner_config.id[-6:]
                    )
                ) from exc
        finally:
            self._runner_config.remove_from_file()

    @utils.start_spinner
    def uninstall(self, **kwargs):
        """Uninstall custom runner."""
        message_queue = kwargs.get("message_queue", queue.Queue)
        message_queue.put(
            MESSAGES.UNINSTALL_RUNNER_MESSAGE.format(
                self._runner_config.name, self._runner_config.id[-6:]
            )
        )

        try:
            self._runner_config.update(force=True)
        except Error:
            print(MESSAGES.LOCAL_UNINSTALL_ACTION_MESSAGE)

        try:
            ongoing_runs = self.get_ongoing_runs
            if ongoing_runs:
                message_queue.put(
                    MESSAGES.DELETE_PODS_MESSAGE.format(
                        self._runner_config.kubectl_namespace
                    )
                )
                for run in ongoing_runs:
                    requests.patch(
                        f"{CONSTS.NEXUS_ENDPOINT}/projects/{run['project_id']}/runs/{run['run_id']}",
                        headers={
                            "Secret-Key": self._runner_config.secret_key,
                            "Content-Type": "application/json",
                        },
                        json={"status": "Cancelled"},
                        timeout=REQUEST_TIME_OUT_SECONDS,
                    )
                self._k8s_batch_api_instance.delete_collection_namespaced_job(
                    self._runner_config.kubectl_namespace
                )

            requests.delete(
                f"{CONSTS.NEXUS_ENDPOINT}/workspaces/ws_{self._runner_config.workspace_id}/runners/{self._runner_config.id}",
                headers={
                    "Secret-Key": self._runner_config.secret_key,
                    "Content-Type": "application/json",
                },
                timeout=REQUEST_TIME_OUT_SECONDS,
            )

            message_queue.put(
                MESSAGES.DELETE_SECRETS_MESSAGE.format(
                    self._runner_config.kubectl_namespace
                )
            )
            self._k8s_core_api_instance.delete_collection_namespaced_secret(
                self._runner_config.kubectl_namespace
            )

            message_queue.put(
                MESSAGES.DELETE_DEPLOYMENTS_MESSAGE.format(
                    self._runner_config.kubectl_namespace
                )
            )
            self._k8s_apps_api_instance.delete_collection_namespaced_deployment(
                self._runner_config.kubectl_namespace
            )

            message_queue.put(
                MESSAGES.DELETE_ROLE_BINDINGS_MESSAGE.format(
                    self._runner_config.kubectl_namespace
                )
            )
            self._k8s_rbac_api_instance.delete_collection_namespaced_role_binding(
                self._runner_config.kubectl_namespace
            )

            message_queue.put(
                MESSAGES.DELETE_ROLES_MESSAGE.format(
                    self._runner_config.kubectl_namespace
                )
            )
            self._k8s_rbac_api_instance.delete_collection_namespaced_role(
                self._runner_config.kubectl_namespace
            )

            message_queue.put(
                MESSAGES.DELETE_SERVICE_ACCOUNTS_MESSAGE.format(
                    self._runner_config.kubectl_namespace
                )
            )
            self._k8s_core_api_instance.delete_collection_namespaced_service_account(
                self._runner_config.kubectl_namespace
            )

            message_queue.put(
                MESSAGES.DELETE_CONFIG_MAPS_MESSAGE.format(
                    self._runner_config.kubectl_namespace
                )
            )
            self._k8s_core_api_instance.delete_collection_namespaced_config_map(
                self._runner_config.kubectl_namespace
            )

            message_queue.put(
                MESSAGES.DELETE_NAMESPACE_MESSAGE.format(
                    self._runner_config.kubectl_namespace
                )
            )
            self._k8s_core_api_instance.delete_namespace(
                self._runner_config.kubectl_namespace
            )

            # Wait for namespace to be fully deleted
            while True:
                try:
                    self._k8s_core_api_instance.read_namespace(
                        self._runner_config.kubectl_namespace
                    )
                    time.sleep(5)
                except k8s.client.rest.ApiException as exc:
                    if exc.status == 404:
                        break
        except k8s.client.rest.ApiException as exc:
            if exc.status != 404:
                raise Error(
                    MESSAGES.UNINSTALL_RUNNER_ERROR_MESSAGE.format(
                        exc.body, self._runner_config.name, self._runner_config.id[-6:]
                    )
                ) from exc
        finally:
            self._runner_config.remove_from_file()

        return MESSAGES.UNINSTALL_RUNNER_SUCCESS_MESSAGE.format(
            self._runner_config.name, self._runner_config.id[-6:]
        )

    @utils.start_spinner
    def suspend(self, **kwargs):
        """Suspend custom runner."""
        if self._runner_config.status.suspended:
            return f"Runner '{self._runner_config.name}' [{self._runner_config.id[-6:]}] has already been suspended."

        message_queue = kwargs.get("message_queue", queue.Queue())
        message_queue.put(
            MESSAGES.SUSPEND_RUNNER_MESSAGE.format(
                self._runner_config.name, self._runner_config.id[-6:]
            )
        )

        try:
            response = requests.patch(
                f"{CONSTS.NEXUS_ENDPOINT}/workspaces/ws_{self._runner_config.workspace_id}/runners/{self._runner_config.id}",
                json={"suspended": True},
                headers={
                    "Secret-Key": self._runner_config.secret_key,
                    "Content-Type": "application/json",
                },
                timeout=REQUEST_TIME_OUT_SECONDS,
            )
            response.raise_for_status()
        except requests.exceptions.ConnectionError:
            print(MESSAGES.LOCAL_SUSPEND_ACTION_MESSAGE)
        except requests.exceptions.HTTPError:
            print(MESSAGES.LOCAL_SUSPEND_ACTION_MESSAGE)

        try:
            self._k8s_apps_api_instance.patch_namespaced_deployment(
                CONSTS.RUNNER_DEPLOYMENT_NAME,
                self._runner_config.kubectl_namespace,
                body={"spec": {"replicas": 0}},
            )
        except k8s.client.rest.ApiException as exc:
            if (
                exc.status != 409
                and json.loads(exc.body).get("reason", "") != "AlreadyExists"
            ):
                raise Error(
                    MESSAGES.SUSPEND_RUNNER_UNKNOWN_ERROR_MESSAGE.format(
                        self._runner_config.name, self._runner_config.id[-6:]
                    )
                ) from exc

        self._runner_config.status.suspended = True
        self._runner_config.write_to_file()
        return MESSAGES.SUSPEND_RUNNER_SUCCESS_MESSAGE.format(
            self._runner_config.name, self._runner_config.id[-6:]
        )

    @utils.start_spinner
    def resume(self, **kwargs):
        """Resume custom runner."""
        if not self._runner_config.status.suspended:
            return f"Runner '{self._runner_config.name}' [{self._runner_config.id[-6:]}] is already active."

        message_queue = kwargs.get("message_queue", queue.Queue())
        message_queue.put(
            MESSAGES.RESUME_RUNNER_MESSAGE.format(
                self._runner_config.name, self._runner_config.id[-6:]
            )
        )

        try:
            utils.setup_docker_registry_credentials(
                self._runner_config, self._k8s_core_api_instance
            )
            response = requests.patch(
                f"{CONSTS.NEXUS_ENDPOINT}/workspaces/ws_{self._runner_config.workspace_id}/runners/{self._runner_config.id}",
                json={"suspended": False},
                headers={
                    "Secret-Key": self._runner_config.secret_key,
                    "Content-Type": "application/json",
                },
                timeout=REQUEST_TIME_OUT_SECONDS,
            )
            response.raise_for_status()
        except requests.exceptions.ConnectionError:
            print(MESSAGES.LOCAL_RESUME_ACTION_MESSAGE)
        except requests.exceptions.HTTPError:
            print(MESSAGES.LOCAL_RESUME_ACTION_MESSAGE)

        try:
            self._k8s_apps_api_instance.patch_namespaced_deployment(
                CONSTS.RUNNER_DEPLOYMENT_NAME,
                self._runner_config.kubectl_namespace,
                body={"spec": {"replicas": 1}},
            )
        except k8s.client.rest.ApiException as exc:
            if (
                exc.status != 409
                and json.loads(exc.body).get("reason", "") != "AlreadyExists"
            ):
                raise Error(
                    MESSAGES.RESUME_RUNNER_UNKNOWN_ERROR_MESSAGE.format(
                        self._runner_config.name, self._runner_config.id[-6:]
                    )
                ) from exc

        self._runner_config.status.suspended = False
        self._runner_config.write_to_file()
        return MESSAGES.RESUME_RUNNER_SUCCESS_MESSAGE.format(
            self._runner_config.name, self._runner_config.id[-6:]
        )

    @utils.start_spinner
    def restart(self, **kwargs):
        """Restart custom runner."""
        message_queue = kwargs.get("message_queue", queue.Queue())
        message_queue.put(
            MESSAGES.RESTART_RUNNER_MESSAGE.format(
                self._runner_config.name, self._runner_config.id[-6:]
            )
        )

        try:
            self._runner_config.update(force=True)
        except Error:
            print(MESSAGES.LOCAL_RESTART_ACTION_MESSAGE)

        self.suspend()
        self.resume()

        try:
            now = datetime.now().isoformat("T") + "Z"
            self._k8s_apps_api_instance.patch_namespaced_deployment(
                CONSTS.RUNNER_DEPLOYMENT_NAME,
                self._runner_config.kubectl_namespace,
                body={
                    "spec": {
                        "template": {
                            "metadata": {
                                "annotations": {
                                    "kubectl.kubernetes.io/restartedAt": now
                                }
                            }
                        }
                    }
                },
            )
        except k8s.client.rest.ApiException as exc:
            if (
                exc.status != 409
                and json.loads(exc.body).get("reason", "") != "AlreadyExists"
            ):
                raise Error(
                    MESSAGES.RESTART_RUNNER_ERROR_MESSAGE.format(
                        exc.body,
                        CONSTS.RUNNER_DEPLOYMENT_NAME,
                        self._runner_config.kubectl_namespace,
                    )
                ) from exc
        return MESSAGES.RESTART_RUNNER_SUCCESS_MESSAGE.format(
            CONSTS.RUNNER_DEPLOYMENT_NAME
        )

    @utils.start_spinner
    def reauth(self, **kwargs):
        """Reauthenticate custom runner."""
        message_queue = kwargs.get("message_queue", queue.Queue())
        new_secret_key = kwargs.get("new_secret_key", "")

        try:
            self._runner_config.secret_key = new_secret_key
            utils.ping_server(self._runner_config)

            # New secret key is valid, save to file
            self._runner_config.write_to_file()

            if self.get_ongoing_runs:
                message_queue.put(
                    MESSAGES.DELETE_PODS_MESSAGE.format(
                        self._runner_config.kubectl_namespace
                    )
                )
                for label_selector in CONSTS.DELETE_JOB_LABEL_SELECTORS:
                    self._k8s_batch_api_instance.delete_collection_namespaced_job(
                        self._runner_config.kubectl_namespace,
                        label_selector=label_selector,
                        propagation_policy="Foreground",
                    )
                while self.get_ongoing_runs:
                    # Wait for all runs to be deleted before continuing
                    time.sleep(5)

        except k8s.client.rest.ApiException as exc:
            raise Error(
                MESSAGES.REAUTH_RUNNER_ERROR_MESSAGE.format(
                    exc.body, self._runner_config.kubectl_namespace
                )
            ) from exc

        utils.apply_config_map(self._runner_config, self._k8s_core_api_instance)
        self.restart()
        return MESSAGES.REAUTH_RUNNER_SUCCESS_MESSAGE.format(
            self._runner_config.name, self._runner_config.id[-6:]
        )

    def dump_logs(self, **kwargs) -> None:
        """Dump custom runner logs."""
        logs_input_dir = CONSTS.RUNNER_LOG_DIR / self._runner_config.id
        logs_output_path = kwargs.get("file", None)
        if logs_output_path:
            logs_output_path = Path(logs_output_path)
            logs_output_path.parent.mkdir(parents=True, exist_ok=True)
        else:
            logs_output_path = Path("/dev/stdout")

        run_hash = kwargs.get("run_hash", None)
        if run_hash:
            run_init_log_suffix = f"{run_hash}-init.log"
            run_log_suffix = f"{run_hash}.log"

            run_init_logs_path = next(
                logs_input_dir.glob(f"*{run_init_log_suffix}"), None
            )
            run_logs_path = next(logs_input_dir.glob(f"*{run_log_suffix}"), None)
            if not run_init_logs_path and not run_logs_path:
                print(MESSAGES.RUN_LOGS_NOT_FOUND_ERROR_MESSAGE.format(run_hash))
                return

            logs = ""
            if run_init_logs_path:
                logs += run_init_logs_path.read_text(encoding="utf-8")
            if run_logs_path:
                logs += run_logs_path.read_text(encoding="utf-8")
            logs_output_path.write_text(logs, encoding="utf-8")
        else:
            runner_logs_path = logs_input_dir / "runner.log"
            if not runner_logs_path.exists():
                print(
                    MESSAGES.RUNNER_LOGS_NOT_FOUND_ERROR_MESSAGE.format(
                        self._runner_config.name, self._runner_config.id[-6:]
                    )
                )
                return

            logs_output_path.write_text(
                runner_logs_path.read_text(encoding="utf-8"), encoding="utf-8"
            )
            if logs_output_path != Path("/dev/stdout"):
                print(
                    MESSAGES.DUMP_RUNNER_LOGS_SUCCESS_MESSAGE.format(
                        logs_output_path.resolve()
                    )
                )

    def display_status(self) -> None:
        """Display custom runner status."""
        try:
            self._runner_config.update()
        except Error as exc:
            if (
                "Workspace ID / Secret Key combination has been invalidated"
                in exc.message
            ):
                self._runner_config.status.valid_secret_key = False
            else:
                print(exc.message)
                sys.exit(1)

        try:
            config_map = self._k8s_core_api_instance.read_namespaced_config_map(
                CONSTS.RUNNER_DEPLOYMENT_CONFIG_NAME,
                self._runner_config.kubectl_namespace,
            ).data

            self._runner_config.name = config_map["runner_name"]
            self._runner_config.id = config_map["runner_id"]
            self._runner_config.secret_key = config_map["secret_key"]
            self._runner_config.workspace_id = config_map["workspace_id"]

            jobs = self._k8s_batch_api_instance.list_namespaced_job(
                self._runner_config.kubectl_namespace
            ).items

        except k8s.client.rest.ApiException as exc:
            raise Error(
                MESSAGES.RETRIEVE_RUNNER_STATUS_ERROR_MESSAGE.format(
                    exc.body, self._runner_config.kubectl_namespace
                )
            ) from exc

        runs_status = []
        if self._runner_config.status.valid_secret_key:
            for job in jobs:
                run_id = job.spec.template.metadata.labels["job-name"].replace(
                    "datature-run-", "run_"
                )
                try:
                    project_id = next(
                        filter(
                            lambda x: x.name == "PROJECT_ID",
                            job.spec.template.spec.containers[0].env,
                        )
                    ).value
                except TypeError:
                    continue

                try:
                    run_response = requests.get(
                        f"{CONSTS.NEXUS_ENDPOINT}/projects/{project_id}/runs/{run_id}",
                        headers={
                            "Secret-Key": self._runner_config.secret_key,
                            "Content-Type": "application/json",
                        },
                        timeout=REQUEST_TIME_OUT_SECONDS,
                    )
                    run_response.raise_for_status()
                except requests.exceptions.ConnectionError as exc:
                    raise Error(
                        MESSAGES.REGISTER_RUNNER_CONNECTION_ERROR_MESSAGE.format(
                            self._runner_config.name, self._runner_config.id[-6:]
                        )
                    ) from exc
                except requests.exceptions.HTTPError as exc:
                    if "ForbiddenError" in run_response.text:
                        raise Error(
                            MESSAGES.REGISTER_RUNNER_FORBIDDEN_ERROR_MESSAGE.format(
                                run_response.text,
                                self._runner_config.name,
                                self._runner_config.id[-6:],
                            )
                        ) from exc
                    if run_response.status_code != 405:
                        raise Error(
                            MESSAGES.REGISTER_RUNNER_UNKNOWN_ERROR_MESSAGE.format(
                                run_response.text,
                                self._runner_config.name,
                                self._runner_config.id[-6:],
                            )
                        ) from exc

                run_image = (
                    job.spec.template.spec.containers[0]
                    .image.split("/")[-1]
                    .split(":")[0]
                )
                run_status = run_response.json()["status"]["overview"]
                run_time_started = job.status.start_time.strftime("%Y-%m-%d %H:%M:%S")
                run_memory_allocated = int(
                    job.spec.template.spec.containers[0].resources.limits.get(
                        "memory", ""
                    )
                )
                run_num_gpus = job.spec.template.spec.containers[
                    0
                ].resources.limits.get("nvidia.com/gpu", "")

                runs_status.append(
                    RunStatus(
                        id=run_id,
                        image_name=run_image,
                        status=run_status,
                        started=run_time_started,
                        allocated_memory=run_memory_allocated,
                        num_gpus=run_num_gpus,
                    )
                )

        self._runner_config.status.runs = runs_status
        utils.format_runner_status(self._runner_config)

        if not self._runner_config.status.valid_secret_key:
            print(
                MESSAGES.RUNNER_INVALID_SECRET_KEY_MESSAGE.format(
                    self._runner_config.workspace_id
                )
            )

        self._runner_config.write_to_file()

    @utils.start_spinner
    def _cleanup(self, message_queue: queue.Queue):
        """Cleanup custom runner resources."""
        message_queue.put(MESSAGES.INSTALL_CLEANUP_MESSAGE)
        self._runner_config.write_to_file()
        return MESSAGES.INSTALL_CLEANUP_SUCCESS_MESSAGE

    @property
    def get_ongoing_runs(self) -> List[dict]:
        """Get ongoing runs in custom runner.

        Returns:
            List[str]: List of ongoing runs.
        """
        try:
            pods = self._k8s_core_api_instance.list_namespaced_pod(
                self._runner_config.kubectl_namespace,
                label_selector="app==datature-trainer",
            ).items
            if len(pods) == 0:
                return []

            return [
                {
                    "run_id": pod.metadata.labels["job-name"].replace(
                        "datature-run-", "run_"
                    ),
                    "project_id": next(
                        filter(
                            lambda x: x.name == "PROJECT_ID",
                            pod.spec.containers[0].env,
                        )
                    ).value,
                }
                for pod in pods
            ]
        except k8s.client.rest.ApiException as exc:
            raise Error(
                MESSAGES.RETRIEVE_RUNS_ERROR_MESSAGE.format(
                    exc.body, self._runner_config.kubectl_namespace
                )
            ) from exc

    @property
    def config(self) -> RunnerConfig:
        """Get custom runner config.

        Returns:
            RunnerConfig: Custom runner config.
        """
        return self._runner_config



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_run.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Run API Test Cases
"""

import unittest
from unittest.mock import ANY, MagicMock

from datature.nexus.api.run import Run
from datature.nexus.client_context import ClientContext

# pylint: disable=W0212


class TestRun(unittest.TestCase):
    """Datature Run API Resource Test Cases."""

    def __init__(self, *args, **kwargs):
        """init global variables."""
        unittest.TestCase.__init__(self, *args, **kwargs)

        context = ClientContext("secret_key", endpoint="http://localhost:8080")
        context.project_id = "project_id"
        context.requester._request = MagicMock()

        self.runs = Run(context)

    def test_list(self):
        """Test list training."""
        self.runs.list()

        self.runs._context.requester._request.assert_called_once_with(
            "GET", "/projects/project_id/runs", ANY, query=None, request_params=None
        )

    def test_retrieve(self):
        """Test retrieve a training."""
        self.runs.get("run_id")

        self.runs._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/runs/run_id",
            ANY,
            query=None,
            request_params=None,
        )

    def test_kill(self):
        """Test kill a training."""
        self.runs.kill("run_id")

        self.runs._context.requester._request.assert_called_once_with(
            "PATCH",
            "/projects/project_id/runs/run_id",
            ANY,
            request_body={"status": "Cancelled"},
            query=None,
            request_params=None,
        )

    def test_start(self):
        """Test start a training."""
        self.runs.start(
            "flow_id",
            {
                "accelerator": {"name": "GPU_T4", "count": 1},
                "checkpoint": {
                    "strategy": "STRAT_LOWEST_VALIDATION_LOSS",
                    "evaluation_interval": 250,
                    "metric": "Loss/total_loss",
                },
                "limit": {"metric": "LIM_MINUTE", "value": 260},
                "preview": True,
                "matrix": True,
            },
        )

        self.runs._context.requester._request.assert_called_once_with(
            "POST",
            "/projects/project_id/runs",
            ANY,
            request_body={
                "flowId": "flow_id",
                "execution": {
                    "accelerator": {"name": "GPU_T4", "count": 1},
                    "checkpoint": {
                        "strategy": "STRAT_LOWEST_VALIDATION_LOSS",
                        "evaluationInterval": 250,
                        "metric": "Loss/total_loss",
                    },
                    "limit": {"metric": "LIM_MINUTE", "value": 260},
                    "debug": False,
                },
                "features": {"preview": True, "matrix": True},
            },
            query=None,
            request_params=None,
        )

    def test_log(self):
        """Test retrieve a training log."""
        self.runs.get_logs("runlog_id")

        self.runs._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/runs/logs/runlog_id",
            ANY,
            query=None,
            request_params=None,
        )

    def test_get_confusion_matrix(self):
        """Test retrieve a training log."""
        self.runs.get_confusion_matrix("run_id")

        self.runs._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/runs/run_id/confusionMatrix",
            ANY,
            query=None,
            request_params=None,
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   frame_fixture.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Upload Session Test Data
"""

# pylint: disable=C0301
register_frame_response = {
    "frame_format": "rectangle",
    "frame_id": "2fa565d3-27e2-4ee6-8362-bb4a3d8e7c12",
    "image_format": "pillow",
    "num_classes": 5,
    "project_id": "b705a30ae26671657f1fd51eb2d4739d",
    "shuffle": True,
    "train_count": 9,
    "validation_count": 419,
    "batch_size": 10,
}



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_requester.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature HTTP Resource Test Cases
"""
# pylint: disable=W0703,W0012,W0212

import json
import unittest

import msgspec
from requests import Session

from datature.nexus import models
from datature.nexus.utils import utils


class TestUtils(unittest.TestCase):
    """Datature Utils Test Cases."""

    def test_find_all_annotations_files(self):
        """Test find_all_annotations_files."""

        # test find all annotations files
        files = utils.find_all_annotations_files("tests/fixture/data")
        assert len(files) == 0

    def test_get_exportable_annotations_formats(self):
        """Test get_exportable_annotations_formats."""

        # test get exportable annotations formats
        formats = utils.get_exportable_annotations_formats("Classification")
        assert len(formats) == 2

        formats = utils.get_exportable_annotations_formats("Keypoint")
        assert len(formats) == 1

        formats = utils.get_exportable_annotations_formats("")
        assert len(formats) == 10

    def test_init_gcs_upload_session(self):
        """Test init_gcs_upload_session."""

        # test init gcs upload session
        session = utils.init_gcs_upload_session()

        assert isinstance(session, Session)

    def test_get_download_path(self):
        """Test get_download_path."""

        # test get download path
        path = utils.get_download_path("../test")
        assert "test" in str(path)

        path = utils.get_download_path()
        assert "datature" in str(path)

    def test_download_files_to_tempfile(self):
        """Test download_files_to_tempfile."""

        # test get download path
        utils.download_files_to_tempfile(
            msgspec.json.decode(
                json.dumps(
                    {
                        "method": "GET",
                        "url": "https://cdn.jsdelivr.net/npm/bootstrap@3.3.7/dist/js/bootstrap.min.js",
                        "expiryDate": 1701755214003,
                    }
                ),
                type=models.DownloadSignedUrl,
            )
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   upload_session_fixture.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Upload Session Test Data
"""

# pylint: disable=C0301
upload_assets_response = {
    "id": "up_057963cb-b945-4fff-9d9e-20adaab82ad5",
    "opId": "op_057963cb-b945-4fff-9d9e-20adaab82ad5",
    "assets": [
        {
            "metadata": {
                "filename": "test.jpeg",
                "mime": "image/jpeg",
                "size": 5613,
                "crc32c": -384617082,
            },
            "upload": {
                "method": "POST",
                "url": "https://storage.googleapis.com/transient.staging.datature.io/593526c7bfd916be7faa8465728f988f/ingestions/37a65f26-3439-472d-af3d-74a4c9767cc6/e42e4eb7-e3a2-4d50-91bf-a829a624b62a/assets/test.jpeg?X-Goog-Algorithm=GOOG4-RSA-SHA256&X-Goog-Credential=gcs-asset-signer%40datature-nexus-staging.iam.gserviceaccount.com%2F45dd6b856ade00802a00dc6d7b4c5ac27a62e1a963fdb7455adf826af85a96ada4ff350daf4c4ea076fa4c18c360296781b66d8a958cb262b45ef1f848fa71c811ec910f8c745f5c46b48c4c41c5dda2c3fce0c702f6558b52872048ddbd2bf752561c7f88eb20346bdc8918b8a8c1b7f45494571399963335286163944c4ee9f31743320ca21c93b0459f3121cd5fc2b5c56232986422acba5a2781127bd738bbf8379e8b7314254922bf76f3773c5193a97acb658a75aabb8bff8ade1c258158d",
                "headers": {
                    "content-type": "image/jpeg",
                    "x-goog-meta-datature-upload": "eyJmaWxlm1haW4iXX0=",
                    "x-goog-hash": "crc32c=6RM1hg==",
                },
                "expiryDate": 1669011530693,
            },
        }
    ],
}



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_file_signature.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Utils Test Cases
"""
# pylint: disable=W0703,W0012,W0212

import unittest

from datature.nexus.utils import file_signature


class TestFileSignature(unittest.TestCase):
    """Datature File Signature Test Cases."""

    def test_jpeg_file(self):
        """Test jpg files."""

        file_mine = file_signature.get_file_mime_by_signature(
            b"\xff\xd8\xff\xe0\x00\x10JFIF\x00\x01\x01\x01\x01,\x01,\x00\x00\xff\xe2\x0cXICC_PROFILE\x00\x01\x01\x00\x00\x0cHLino\x02\x10\x00\x00mntrRGB XYZ \x07\xce\x00\x02\x00\t\x00\x06\x001\x00\x00acspMSFT\x00\x00\x00\x00IEC sRGB\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\xf6\xd6\x00\x01\x00\x00\x00\x00\xd3-HP  \x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x11cprt\x00\x00\x01P\x00\x00\x003desc\x00\x00\x01\x84\x00\x00\x00lwtpt\x00\x00\x01\xf0\x00\x00\x00\x14bkpt\x00\x00\x02\x04\x00\x00\x00\x14rXYZ\x00\x00\x02\x18\x00\x00\x00\x14gXYZ\x00\x00\x02,\x00\x00\x00\x14bXYZ\x00\x00\x02@\x00\x00\x00\x14dmnd\x00\x00\x02T\x00\x00\x00pdmdd\x00\x00\x02\xc4\x00\x00\x00\x88vued\x00\x00\x03L\x00\x00\x00\x86view\x00\x00\x03\xd4\x00\x00\x00$lumi\x00\x00\x03\xf8\x00\x00\x00\x14meas\x00\x00\x04\x0c\x00\x00\x00$tech\x00\x00\x040\x00\x00\x00\x0crTRC\x00\x00\x04<\x00\x00\x08\x0cgTRC\x00\x00\x04<\x00\x00\x08\x0cbTRC\x00\x00\x04<\x00\x00\x08\x0ctext\x00\x00\x00\x00Copyright (c) 1998"
        )

        assert file_mine == "image/jpeg"

    def test_not_supported_file(self):
        """Test not supported files."""

        file_mine = file_signature.get_file_mime_by_signature(
            b"%PDF-1.4\n%\xd3\xeb\xe9\xe1\n1 0 obj\n<</Creator (Mozilla/5.0 \\(X11; Linux x86_64\\) AppleWebKit/537.36 \\(KHTML, like Gecko\\) HeadlessChrome/117.0.0.0 Safari/537.36)\n/Producer (Skia/PDF m117)\n/CreationDate (D:20231003074453+00'00')\n/ModDate (D:20231003074453+00'00')>>\nendobj\n3 0 obj\n<</ca 1\n/BM /Normal>>\nendobj\n6 0 obj\n<</Filter /FlateDecode\n/Length 707>> stream\nx\x9c\x95U\xdbj\x1b1\x10}\xdf\xaf\xd0s\xa1\x8af4\xd2\xcc@\x08\xd4\xce\x85>\x04\xdab\xe8\x07\xf4\x12(\xa4\x90\xf4\xff\xa1g\xd7\xb2w\xad\x8d\x8bk\xc3\x12\xef\xc9"
        )

        assert file_mine is None
